import logging
import os
import pickle
from pathlib import Path
from typing import Optional, Union

from .. import model as m
from ..sqlalchemy_ import sa
from ...mapper.model_object import Identifier

logger = logging.getLogger(
    __name__
)


def _parent_identifier(
        directory: str
) -> Optional[str]:
    """
    Read the parent identifier for a fit in a directory.

    Defaults to None if no .parent_identifier file is found.
    """
    try:
        with open(f"{directory}/.parent_identifier") as f:
            return f.read()
    except FileNotFoundError:
        return None


class Scraper:
    def __init__(
            self,
            directory: Union[Path, str],
            session: sa.orm.Session
    ):
        """
        Facilitates scraping of data output into a directory
        into the database.

        Parameters
        ----------
        directory
            A directory in which data has been stored
        session
            A database session
        """
        self.directory = directory
        self.session = session

    def scrape(self):
        """
        Recursively scrape fits from the directory and
        add them to the session
        """
        for fit in self._fits():
            self.session.add(
                fit
            )
        for grid_search in self._grid_searches():
            self.session.add(
                grid_search
            )

    def _fits(self):
        """
        Scrape data output into a directory tree so it can be added to the
        aggregator database.

        Returns
        -------
        Generator yielding Fit database objects
        """
        logger.info(
            f"Scraping directory {self.directory}"
        )
        from autofit.aggregator.aggregator import Aggregator as ClassicAggregator
        aggregator = ClassicAggregator(
            self.directory
        )
        logger.info(
            f"{len(aggregator)} searches found"
        )
        for item in aggregator:

            is_complete = os.path.exists(
                f"{item.directory}/.completed"
            )

            parent_identifier = _parent_identifier(
                directory=item.directory
            )

            model = item.model
            samples = item.samples

            identifier = _make_identifier(item)

            logger.info(
                f"Creating fit for: "
                f"{item.search.paths.path_prefix} "
                f"{item.search.unique_tag} "
                f"{item.search.name} "
                f"{identifier} ")

            try:
                instance = samples.max_log_likelihood_instance
            except (AttributeError, NotImplementedError):
                instance = None

            try:
                fit = self._retrieve_model_fit(
                    item
                )
                logger.warning(
                    f"Fit already existed with identifier {identifier}"
                )
            except sa.orm.exc.NoResultFound:
                try:
                    log_likelihood = samples.max_log_likelihood_sample.log_likelihood
                except AttributeError:
                    log_likelihood = None
                fit = m.Fit(
                    id=identifier,
                    name=item.search.name,
                    unique_tag=item.search.unique_tag,
                    model=model,
                    instance=instance,
                    is_complete=is_complete,
                    info=item.info,
                    max_log_likelihood=log_likelihood,
                    parent_id=parent_identifier
                )

            pickle_path = Path(item.pickle_path)
            _add_pickles(
                fit,
                pickle_path
            )

            yield fit

    def _grid_searches(
            self
    ):
        """
        Retrieve grid searches recursively from an output directory by
        searching for the .is_grid_search file.

        Should be called after adding Fits as it relies on querying fits

        Yields
        ------
        Fit objects representing grid searches with child fits associated
        """
        from autofit.aggregator.aggregator import Aggregator as ClassicAggregator
        for root, _, filenames in os.walk(self.directory):
            if ".is_grid_search" in filenames:
                path = Path(root)

                is_complete = (path / ".completed").exists()

                with open(
                        path / ".is_grid_search"
                ) as f:
                    unique_tag = f.read()

                grid_search = m.Fit(
                    id=path.name,
                    unique_tag=unique_tag,
                    is_grid_search=True,
                    parent_id=_parent_identifier(
                        root
                    ),
                    is_complete=is_complete
                )

                pickle_path = path / "pickles"
                _add_pickles(
                    grid_search,
                    pickle_path
                )

                aggregator = ClassicAggregator(
                    root
                )
                for item in aggregator:
                    fit = self._retrieve_model_fit(
                        item
                    )
                    grid_search.children.append(
                        fit
                    )
                yield grid_search

    def _retrieve_model_fit(
            self,
            item
    ) -> m.Fit:
        """
        Retrieve a Fit, if one exists, corresponding to a given SearchOutput

        Parameters
        ----------
        item
            A SearchOutput from the classic Aggregator

        Returns
        -------
        A fit with the corresponding identifier

        Raises
        ------
        NoResultFound
            If no fit is found with the identifier
        """
        return self.session.query(
            m.Fit
        ).filter(
            m.Fit.id == _make_identifier(
                item
            )
        ).one()


def _make_identifier(
        item
) -> str:
    """
    Create a unique identifier for a SearchOutput.

    This accounts for the Search, Model and unique_tag

    Parameters
    ----------
    item
        An output from the classic aggregator

    Returns
    -------
    A unique identifier that is sensitive to changes that affect
    the search
    """
    search = item.search
    model = item.model
    return str(Identifier([
        search,
        model,
        search.unique_tag
    ]))


def _add_pickles(
        fit: m.Fit,
        pickle_path: Path
):
    """
    Load pickles from the path and add them to the database.

    Parameters
    ----------
    fit
        A fit to which the pickles belong
    pickle_path
        The path in which the pickles are stored
    """
    try:
        filenames = os.listdir(
            pickle_path
        )
    except FileNotFoundError as e:
        logger.exception(e)
        filenames = []

    for filename in filenames:

        try:
            with open(
                    pickle_path / filename,
                    "r+b"
            ) as f:
                fit[
                    filename.split(".")[0]
                ] = pickle.load(f)
        except (pickle.UnpicklingError, ModuleNotFoundError) as e:

            if filename == "dynesty.pickle":
                continue

            raise pickle.UnpicklingError(f"Failed to unpickle: {pickle_path} {filename}") from e



import inspect
from collections.abc import Iterable
from hashlib import md5
from typing import Optional

from autoconf import conf
# floats are rounded to this increment so floating point errors
# have no impact on identifier value
from autoconf.class_path import get_class_path

RESOLUTION = 1e-8


class IdentifierField:
    """
    A field that the identifier depends on.

    If the value of the field is changed then the identifier
    must be recomputed prior to use.
    """

    def __set_name__(
            self,
            owner: object,
            name: str
    ):
        """
        Called on instantiation

        Parameters
        ----------
        owner
            The object for which this field is an attribute
        name
            The name of the attribute
        """
        self.private = f"_{name}"
        setattr(
            owner,
            self.private,
            None
        )

    def __get__(
            self,
            obj: object,
            objtype=None
    ) -> Optional:
        """
        Retrieve the value of this field.

        Parameters
        ----------
        obj
            The object for which this field is an attribute
        objtype

        Returns
        -------
        The value (or None if it has not been set)
        """
        return getattr(
            obj,
            self.private
        )

    def __set__(self, obj, value):
        """
        Set a value for this field

        Parameters
        ----------
        obj
            The object for which the field is an attribute
        value
            A new value for the attribute
        """
        obj._identifier = None
        setattr(
            obj,
            self.private,
            value
        )


class Identifier:
    def __init__(self, obj, version=None):
        """
        Wraps an object and recursively generates an identifier

        The version can be set in general.ini (output/identifier_version). It can
        be overridden by specifying it explicitly in the constructor.

        The version determines how the identifier is generated.

        Version History
        ---------------
        1 - Original version
        2 - Accounts for the class of objects
        3 - Include class path to distinguish prior models
        4 - __exclude_identifier_fields__ can be used to exclude specific fields

        """
        self._identifier_version = version or conf.instance[
            "general"
        ][
            "output"
        ][
            "identifier_version"
        ]

        self.hash_list = list()
        self._add_value_to_hash_list(
            obj
        )

    @property
    def description(self):
        return "\n".join(self.hash_list)

    def _add_value_to_hash_list(self, value):
        """
        Add some object and recursively add its children to the hash_list.

        The md5 hash of this object is taken to create an identifier.

        If an object specifies __identifier_fields__ then only attributes
        with a name in this list are included.

        Parameters
        ----------
        value
            An object
        """
        from .model_object import ModelObject
        if self._identifier_version >= 3:
            if inspect.isclass(value):
                self.add_value_to_hash_list(
                    get_class_path(value)
                )
                return

        if isinstance(
                value, Exception
        ):
            raise value
        if hasattr(value, "__dict__"):
            if self._identifier_version >= 2:
                if hasattr(value, "__class__"):
                    self.add_value_to_hash_list(
                        value.__class__.__name__
                    )
            d = value.__dict__

            if hasattr(
                    value,
                    "__identifier_fields__"
            ):
                fields = value.__identifier_fields__

                try:
                    d = {
                        k: getattr(value, k)
                        for k in fields
                    }
                except AttributeError as e:
                    raise AssertionError(
                        f"Missing identifier fields for {type(value)}"
                    ) from e
            elif hasattr(
                    value,
                    "__class__"
            ) and not inspect.isclass(
                value
            ) and not isinstance(
                value,
                ModelObject
            ):
                args = inspect.getfullargspec(
                    value.__class__
                ).args
                d = {
                    k: v
                    for k, v
                    in d.items()
                    if k in args
                }
                if self._identifier_version >= 4 and hasattr(
                        value,
                        "__exclude_identifier_fields__"
                ):
                    excluded_fields = value.__exclude_identifier_fields__

                    d = {
                        k: v
                        for k, v
                        in d.items()
                        if k not in excluded_fields
                    }
            self.add_value_to_hash_list(
                d
            )
        elif isinstance(
                value, dict
        ):
            for key, value in value.items():
                if not (key.startswith("_") or key in ("id", "paths")):
                    self.hash_list.append(key)
                    self.add_value_to_hash_list(
                        value
                    )
        elif isinstance(
                value, float
        ):
            try:
                value = RESOLUTION * round(
                    value / RESOLUTION
                )
            except OverflowError:
                pass

            self.hash_list.append(
                str(value)
            )
        elif isinstance(
                value,
                (str, int, bool)
        ):
            self.hash_list.append(
                str(value)
            )
        elif isinstance(value, Iterable):
            for value in value:
                self.add_value_to_hash_list(
                    value
                )

    def add_value_to_hash_list(self, value):
        if isinstance(
                value,
                property
        ):
            return
        self._add_value_to_hash_list(
            value
        )

    def __str__(self):
        return md5(".".join(
            self.hash_list
        ).encode("utf-8")).hexdigest()

    def __repr__(self):
        return f"<{self.__class__.__name__} {self}>"

    def __eq__(self, other):
        return str(self) == str(other)



import copy
import csv
import logging
from os import path
from typing import List, Tuple, Union, Type, Optional, Dict

from autofit import exc
from autofit.mapper import prior as p
from autofit.non_linear.abstract_search import NonLinearSearch
from autofit.non_linear.parallel import Process
from autofit.text.text_util import padding
from .job import Job
from .result import GridSearchResult
from .result_builder import ResultBuilder


class Sequential:
    # noinspection PyUnusedLocal
    @staticmethod
    def run_jobs(jobs, *_, **kwargs):
        for job_ in jobs:
            yield job_.perform()


class GridSearch:

    def __init__(
            self,
            search,
            number_of_steps: int = 4,
            number_of_cores: int = 1,
            result_output_interval: int = 100,
    ):
        """
        Performs a non linear optimiser search for each square in a grid. The dimensionality of the search depends on
        the number of distinct priors passed to the fit function. (1 / step_size) ^ no_dimension steps are performed
        per an optimisation.

        Parameters
        ----------
        number_of_steps
            The number of steps to go in each direction
        search
            The class of the search that is run at each step
        result_output_interval
            The interval between saving a GridSearchResult object via paths
        """
        self.number_of_steps = number_of_steps
        self.search = search
        self.prior_passer = search.prior_passer

        self._logger = None

        self.logger.info(
            "Creating grid search"
        )

        self.number_of_cores = number_of_cores or 1

        self.logger.info(
            f"Using {number_of_cores} core(s)"
        )

        self._result_output_interval = result_output_interval

    __exclude_identifier_fields__ = ("number_of_cores",)

    def __getstate__(self):
        """
        Remove the logger for pickling
        """
        state = self.__dict__.copy()
        if "_logger" in state:
            del state["_logger"]
        return state

    @property
    def logger(self):
        if not hasattr(self, "_logger") or self._logger is None:
            self._logger = logging.getLogger(
                f"GridSearch ({self.search.name})"
            )
        return self._logger

    @property
    def paths(self):
        return self.search.paths

    @property
    def parallel(self) -> bool:
        """
        Should this grid search run in parallel?
        """
        return self.number_of_cores > 1

    @property
    def step_size(self):
        """
        Returns
        -------
        step_size: float
            The size of a step in any given dimension in hyper space.
        """
        return 1 / self.number_of_steps

    def make_physical_lists(self, grid_priors) -> List[List[float]]:
        lists = self.make_lists(grid_priors)
        return [
            [prior.value_for(value) for prior, value in zip(grid_priors, l)]
            for l in lists
        ]

    def make_lists(self, grid_priors):
        """
        Produces a list of lists of floats, where each list of floats represents the values in each dimension for one
        step of the grid search.

        Parameters
        ----------
        grid_priors: [p.Prior]
            A list of priors that are to be searched using the grid search.

        Returns
        -------
        lists: [[float]]
        """
        return make_lists(
            len(grid_priors), step_size=self.step_size, centre_steps=False
        )

    def make_arguments(self, values, grid_priors):
        arguments = {}
        for value, grid_prior in zip(values, grid_priors):
            if (
                    float("-inf") == grid_prior.lower_limit
                    or float("inf") == grid_prior.upper_limit
            ):
                raise exc.PriorException(
                    "Priors passed to the grid search must have definite limits"
                )
            lower_limit = grid_prior.lower_limit + value * grid_prior.width
            upper_limit = (
                    grid_prior.lower_limit
                    + (value + self.step_size) * grid_prior.width
            )
            prior = p.UniformPrior(lower_limit=lower_limit, upper_limit=upper_limit)
            arguments[grid_prior] = prior
        return arguments

    def model_mappers(self, model, grid_priors):
        grid_priors = model.sort_priors_alphabetically(set(grid_priors))
        lists = self.make_lists(grid_priors)
        for values in lists:
            arguments = self.make_arguments(values, grid_priors)
            yield model.mapper_from_partial_prior_arguments(arguments)

    def fit(
            self,
            model,
            analysis,
            grid_priors,
            info: Optional[Dict] = None,
            parent: Optional[NonLinearSearch] = None
    ):
        """
        Fit an analysis with a set of grid priors. The grid priors are priors associated with the model mapper
        of this instance that are replaced by uniform priors for each step of the grid search.

        Parameters
        ----------
        parent
            Optionally specify a parent, for example a search performed prior
            to this grid search
        model
        analysis: autofit.non_linear.non_linear.Analysis
            An analysis used to determine the fitness of a given model instance
        grid_priors: [p.Prior]
            A list of priors to be substituted for uniform priors across the grid.

        Returns
        -------
        result: GridSearchResult
            An object that comprises the results from each individual fit
        """
        self.paths.model = model
        self.paths.search = self
        if parent is not None:
            self.paths.parent = parent.paths

        self.logger.info(
            "Running grid search..."
        )
        process_class = Process if self.parallel else Sequential
        # noinspection PyArgumentList
        return self._fit(
            model=model,
            analysis=analysis,
            grid_priors=grid_priors,
            process_class=process_class,
            info=info
        )

    def _fit(
            self,
            model,
            analysis,
            grid_priors,
            process_class=Union[Type[Process], Type[Sequential]],
            info: Optional[Dict] = None
    ):
        """
        Perform the grid search in parallel, with all the optimisation for each grid square being performed on a
        different process.

        Parameters
        ----------
        model
        analysis
            An analysis
        grid_priors
            Priors describing the position in the grid
        process_class
            Class used to dispatch jobs

        Returns
        -------
        result: GridSearchResult
            The result of the grid search
        """
        self.logger.info(
            "...in parallel"
        )

        grid_priors = model.sort_priors_alphabetically(
            set(grid_priors)
        )
        lists = self.make_lists(grid_priors)

        builder = ResultBuilder(
            lists=lists,
            grid_priors=grid_priors
        )

        self.save_metadata()

        def save_results():
            self.paths.save_object(
                "result",
                builder()
            )

        def write_results():
            self.logger.debug(
                "Writing results"
            )
            with open(path.join(self.paths.output_path, "results.csv"), "w+") as f:
                writer = csv.writer(f)
                writer.writerow([
                    ["index"]
                    + list(map(model.name_for_prior, grid_priors))
                    + ["likelihood_merit"]
                ])

                for results in results_list:
                    writer.writerow([
                        padding(f"{value:.2f}")
                        for value in results
                    ])

        results_list = []

        for i, job_result in enumerate(
                process_class.run_jobs(
                    self.make_jobs(
                        model,
                        analysis,
                        grid_priors,
                        info
                    ),
                    self.number_of_cores
                )
        ):
            builder.add(
                job_result
            )
            results_list.append(job_result.result_list_row)
            write_results()
            if i % self._result_output_interval == 0:
                save_results()

        save_results()
        self.paths.completed()

        return builder()

    def save_metadata(self):
        self.paths.save_parent_identifier()
        self.paths.save_unique_tag(
            is_grid_search=True
        )

    def make_jobs(self, model, analysis, grid_priors, info: Optional[Dict] = None):
        grid_priors = model.sort_priors_alphabetically(
            set(grid_priors)
        )
        lists = self.make_lists(grid_priors)

        jobs = list()

        for index, values in enumerate(lists):
            jobs.append(
                self.job_for_analysis_grid_priors_and_values(
                    analysis=copy.deepcopy(analysis),
                    model=model,
                    grid_priors=grid_priors,
                    values=values,
                    index=index,
                    info=info
                )
            )
        return jobs

    def job_for_analysis_grid_priors_and_values(
            self,
            model,
            analysis,
            grid_priors,
            values,
            index,
            info: Optional[Dict] = None
    ):
        arguments = self.make_arguments(values=values, grid_priors=grid_priors)
        model = model.mapper_from_partial_prior_arguments(arguments=arguments)

        labels = []
        for prior in sorted(arguments.values(), key=lambda pr: pr.id):
            labels.append(
                "{}_{:.2f}_{:.2f}".format(
                    model.name_for_prior(prior), prior.lower_limit, prior.upper_limit
                )
            )

        name_path = path.join(
            self.paths.name,
            self.paths.identifier,
            "_".join(labels),
        )

        search_instance = self.search_instance(name_path=name_path)
        search_instance.paths.model = model

        return Job(
            search_instance=search_instance,
            model=model,
            analysis=analysis,
            arguments=arguments,
            index=index,
            info=info
        )

    def search_instance(self, name_path):

        search_instance = self.search.copy_with_paths(
            self.paths.create_child(
                name=name_path,
                path_prefix=self.paths.path_prefix,
                is_identifier_in_paths=False
            )
        )

        for key, value in self.__dict__.items():
            if key not in ("model", "instance", "paths", "search"):
                try:
                    setattr(search_instance, key, value)
                except AttributeError:
                    pass

        search_instance.number_of_cores = self.number_of_cores

        if self.number_of_cores > 1:

            search_instance.number_of_cores = 1

        return search_instance


def grid(fitness_function, no_dimensions, step_size):
    """
    Grid2D search using a fitness function over a given number of dimensions and a given step size between inclusive
    limits of 0 and 1.

    Parameters
    ----------
    fitness_function: function
        A function that takes a tuple of floats as an argument
    no_dimensions: int
        The number of dimensions of the grid search
    step_size: float
        The step size of the grid search

    Returns
    -------
    best_arguments: tuple[float]
        The tuple of arguments that gave the highest fitness
    """
    best_fitness = float("-inf")
    best_arguments = None

    for arguments in make_lists(no_dimensions, step_size):
        fitness = fitness_function(tuple(arguments))
        if fitness > best_fitness:
            best_fitness = fitness
            best_arguments = tuple(arguments)

    return best_arguments


def make_lists(
        no_dimensions: int,
        step_size: Union[Tuple[float], float],
        centre_steps=True
):
    """
        Returns a list of lists of floats covering every combination across no_dimensions of points of integer step size
    between 0 and 1 inclusive.

    Parameters
    ----------
    no_dimensions
        The number of dimensions, that is the length of the lists
    step_size
        The step size. This can be a float or a tuple with the same number of dimensions
    centre_steps

    Returns
    -------
    lists: [[float]]
        A list of lists
    """
    if isinstance(step_size, float):
        step_size = tuple(
            step_size
            for _
            in range(no_dimensions)
        )

    if no_dimensions == 0:
        return [[]]

    sub_lists = make_lists(
        no_dimensions - 1,
        step_size[1:],
        centre_steps=centre_steps
    )
    step_size = step_size[0]
    return [
        [
            step_size * value + (
                0.5 * step_size
                if centre_steps
                else 0)
        ] + sub_list
        for value in range(int((1 / step_size)))
        for sub_list in sub_lists
    ]



from os import path
from typing import Optional

import numpy as np
import zeus
from autofit.database.sqlalchemy_ import sa

from autoconf import conf
from autofit import exc
from autofit.mapper.model_mapper import ModelMapper
from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.non_linear.mcmc.abstract_mcmc import AbstractMCMC
from autofit.non_linear.mcmc.auto_correlations import AutoCorrelationsSettings
from autofit.non_linear.mcmc.zeus.samples import ZeusSamples
from autofit.non_linear.abstract_search import PriorPasser
from autofit.non_linear.initializer import Initializer
from autofit.non_linear.mcmc.zeus.plotter import ZeusPlotter
from autofit.plot.output import Output


class Zeus(AbstractMCMC):
    __identifier_fields__ = (
        "nwalkers",
        "tune",
        "tolerance",
        "patience",
        "mu",
        "light_mode"
    )

    def __init__(
            self,
            name: Optional[str] = None,
            path_prefix: Optional[str] = None,
            unique_tag: Optional[str] = None,
            prior_passer: Optional[PriorPasser] = None,
            initializer: Optional[Initializer] = None,
            auto_correlations_settings=AutoCorrelationsSettings(),
            iterations_per_update: int = None,
            number_of_cores: int = None,
            session: Optional[sa.orm.Session] = None,
            **kwargs
    ):
        """
        An Zeus non-linear search.

        For a full description of Zeus, checkout its Github and readthedocs webpages:

        https://github.com/minaskar/zeus

        https://zeus-mcmc.readthedocs.io/en/latest/

        If you use `Zeus` as part of a published work, please cite the package following the instructions under the
        *Attribution* section of the GitHub page.

        Parameters
        ----------
        name
            The name of the search, controlling the last folder results are output.
        path_prefix
            The path of folders prefixing the name folder where results are output.
        unique_tag
            The name of a unique tag for this model-fit, which will be given a unique entry in the sqlite database
            and also acts as the folder after the path prefix and before the search name.
        prior_passer
            Controls how priors are passed from the results of this `NonLinearSearch` to a subsequent non-linear search.
        nwalkers
            The number of walkers in the ensemble used to sample parameter space.
        nsteps
            The number of steps that must be taken by every walker. The `NonLinearSearch` will thus run for nwalkers *
            nsteps iterations.
        initializer
            Generates the initialize samples of non-linear parameter space (see autofit.non_linear.initializer).
        auto_correlations_settings : AutoCorrelationsSettings
            Customizes and performs auto correlation calculations performed during and after the search.
        number_of_cores
            The number of cores Zeus sampling is performed using a Python multiprocessing Pool instance. If 1, a
            pool instance is not created and the job runs in serial.
        session
            An SQLalchemy session instance so the results of the model-fit are written to an SQLite database.
        """

        super().__init__(
            name=name,
            path_prefix=path_prefix,
            unique_tag=unique_tag,
            prior_passer=prior_passer,
            initializer=initializer,
            auto_correlations_settings=auto_correlations_settings,
            iterations_per_update=iterations_per_update,
            session=session,
            **kwargs
        )

        self.number_of_cores = number_of_cores or self._config("parallel", "number_of_cores")

        self.logger.debug("Creating Zeus Search")

    class Fitness(AbstractMCMC.Fitness):
        def __call__(self, parameters):
            try:
                return self.figure_of_merit_from(parameter_list=parameters)
            except exc.FitException:
                return self.resample_figure_of_merit

        def figure_of_merit_from(self, parameter_list):
            """
            The figure of merit is the value that the `NonLinearSearch` uses to sample parameter space. 
            
            `Zeus` uses the log posterior.
            """
            log_posterior = self.log_posterior_from(parameter_list=parameter_list)

            if np.isnan(log_posterior):
                return self.resample_figure_of_merit

            return log_posterior

    def _fit(self, model: AbstractPriorModel, analysis, log_likelihood_cap=None):
        """
        Fit a model using Zeus and the Analysis class which contains the data and returns the log likelihood from
        instances of the model, which the `NonLinearSearch` seeks to maximize.

        Parameters
        ----------
        model : ModelMapper
            The model which generates instances for different points in parameter space.
        analysis : Analysis
            Contains the data and the log likelihood function which fits an instance of the model to the data, returning
            the log likelihood the `NonLinearSearch` maximizes.

        Returns
        -------
        A result object comprising the Samples object that inclues the maximum log likelihood instance and full
        chains used by the fit.
        """

        pool = self.make_pool()

        fitness_function = self.fitness_function_from_model_and_analysis(
            model=model, analysis=analysis
        )

        if self.paths.is_object("zeus"):

            zeus_sampler = self.zeus_pickled

            zeus_state = zeus_sampler.get_last_sample()
            log_posterior_list = zeus_sampler.get_last_log_prob()

            samples = self.samples_from(model=model)

            total_iterations = zeus_sampler.iteration

            if samples.converged:
                iterations_remaining = 0
            else:
                iterations_remaining = self.config_dict_run["nsteps"] - total_iterations

                self.logger.info("Existing Zeus samples found, resuming non-linear search.")

        else:

            zeus_sampler = zeus.EnsembleSampler(
                nwalkers=self.config_dict_search["nwalkers"],
                ndim=model.prior_count,
                logprob_fn=fitness_function.__call__,
                pool=pool,
            )

            zeus_sampler.ncall_total = 0

            unit_parameter_lists, parameter_lists, log_posterior_list = self.initializer.samples_from_model(
                total_points=zeus_sampler.nwalkers,
                model=model,
                fitness_function=fitness_function,
            )

            zeus_state = np.zeros(shape=(zeus_sampler.nwalkers, model.prior_count))

            self.logger.info("No Zeus samples found, beginning new non-linear search.")

            for index, parameters in enumerate(parameter_lists):

                zeus_state[index, :] = np.asarray(parameters)

            total_iterations = 0
            iterations_remaining = self.config_dict_run["nsteps"]

        while iterations_remaining > 0:

            if self.iterations_per_update > iterations_remaining:
                iterations = iterations_remaining
            else:
                iterations = self.iterations_per_update

            for sample in zeus_sampler.sample(
                    start=zeus_state,
                    log_prob0=log_posterior_list,
                    iterations=iterations,
                    progress=True,
            ):

                pass

            zeus_sampler.ncall_total += zeus_sampler.ncall

            self.paths.save_object(
                "zeus",
                zeus_sampler
            )

            zeus_state = zeus_sampler.get_last_sample()
            log_posterior_list = zeus_sampler.get_last_log_prob()

            total_iterations += iterations
            iterations_remaining = self.config_dict_run["nsteps"] - total_iterations

            samples = self.perform_update(
                model=model, analysis=analysis, during_analysis=True
            )

            if self.auto_correlations_settings.check_for_convergence:
                if zeus_sampler.iteration > self.auto_correlations_settings.check_size:
                    if samples.converged:
                        iterations_remaining = 0

            auto_correlation_time = zeus.AutoCorrTime(samples=zeus_sampler.get_chain())

            discard = int(3.0 * np.max(auto_correlation_time))
            thin = int(np.max(auto_correlation_time) / 2.0)
            chain = zeus_sampler.get_chain(discard=discard, thin=thin, flat=True)

            if "maxcall" in self.kwargs:
                if zeus_sampler.ncall_total > self.kwargs["maxcall"]:
                    iterations_remaining = 0

        self.logger.info("Zeus sampling complete.")

    def fitness_function_from_model_and_analysis(self, model, analysis, log_likelihood_cap=None):

        return Zeus.Fitness(
            paths=self.paths,
            model=model,
            analysis=analysis,
            samples_from_model=self.samples_from,
            log_likelihood_cap=log_likelihood_cap
        )

    def samples_from(self, model):
        """Create a `Samples` object from this non-linear search's output files on the hard-disk and model.

        Parameters
        ----------
        model
            The model which generates instances for different points in parameter space. This maps the points from unit
            cube values to physical values via the priors.
        paths : af.Paths
            Manages all paths, e.g. where the search outputs are stored, the `NonLinearSearch` chains,
            etc.
        """

        return ZeusSamples.from_results_internal(
            results_internal=self.zeus_pickled,
            model=model,
            auto_correlation_settings=self.auto_correlations_settings,
            time=self.timer.time
        )

    @property
    def zeus_pickled(self):
        return self.paths.load_object(
            "zeus"
        )

    def plot_results(self, samples):

        def should_plot(name):
            return conf.instance["visualize"]["plots_search"]["emcee"][name]

        plotter = ZeusPlotter(
            samples=samples,
            output=Output(path=path.join(self.paths.image_path, "search"), format="png")
        )

        if should_plot("corner"):
            plotter.corner()

        if should_plot("trajectories"):
            plotter.trajectories()

        if should_plot("likelihood_series"):
            plotter.likelihood_series()

        if should_plot("time_series"):
            plotter.time_series()





from autofit.database.migration import Step, Migrator

steps = [
    Step(
        "ALTER TABLE fit ADD name VARCHAR;",
        "ALTER TABLE fit ADD path_prefix VARCHAR;"
    ),
    Step(
        "CREATE TABLE named_instance (id INTEGER NOT NULL , name VARCHAR, fit_id VARCHAR, PRIMARY KEY (id), FOREIGN KEY (fit_id) REFERENCES fit (id));"
    ),
    Step(
        "ALTER TABLE fit ADD max_log_likelihood FLOAT;"
    )
]

migrator = Migrator(*steps)



from .condition import NameCondition, ValueCondition, StringValueCondition, TypeCondition
from .junction import And, Or
from .query import NamedQuery, Attribute, BooleanAttribute, ChildQuery
from .query.info import *


class N(NameCondition):
    pass


class V(ValueCondition):
    pass


class SV(StringValueCondition):
    pass


class T(TypeCondition):
    pass


class Q(NamedQuery):
    pass


class A(Attribute):
    pass


class BA(BooleanAttribute):
    pass



from abc import ABC

from autoconf import conf
from autofit import exc
from autofit.non_linear.abstract_search import NonLinearSearch



class AbstractOptimizer(NonLinearSearch, ABC):
    @property
    def config_type(self):
        return conf.instance["non_linear"]["optimize"]

    class Fitness(NonLinearSearch.Fitness):

        def __call__(self, parameters):
            try:
                return self.figure_of_merit_from(parameter_list=parameters)
            except exc.FitException:
                return self.resample_figure_of_merit

        def figure_of_merit_from(self, parameter_list):
            """
            The figure of merit is the value that the `NonLinearSearch` uses to sample parameter space. `Emcee`
            uses the log posterior.

            L-BFGS uses the chi-squared value, which is the -2.0*log_posterior.
            """
            return -2.0 * self.log_posterior_from(parameter_list=parameter_list)

    def fitness_function_from_model_and_analysis(self, model, analysis, log_likelihood_cap=None):

        return self.Fitness(
            paths=self.paths,
            model=model,
            analysis=analysis,
            samples_from_model=self.samples_from,
            log_likelihood_cap=log_likelihood_cap
        )


from abc import ABC, abstractmethod
from collections import defaultdict
from typing import Set, Iterable

from .condition import AbstractCondition, Table, none_table


class AbstractJunction(AbstractCondition, ABC):
    def __new__(
            cls,
            *conditions: AbstractCondition
    ):
        """
        If only a single extant condition is passed in then that
        condition should simply be returned.
        """
        conditions = cls._match_conditions(conditions)
        if len(conditions) == 1:
            return list(conditions)[0]
        return object.__new__(cls)

    def __init__(
            self,
            *conditions: AbstractCondition
    ):
        """
        A combination of two or more conditions. A set of rules allow
        the query to be expressed in the simplest terms possible

        Any subjunctions of the same type are unwrapped.
        i.e. And(A, B) & And(C, D) -> And(A, B, C, D)

        Conditions are a set so duplication is removed.
        i.e. And(A, B, A) -> And(A, B)

        NamedQuery child conditions are matched by their name creating
        one NameQuery for each name and applying the junction type to
        their conditions.
        i.e. And(Named('name', A), Named('name', B)) -> Named('name', And(A, B))

        An AttributeQuery is on an attribute of the fit class. If an AttributeQuery
        is present in a conjunction then the fit_query must be returned instead
        of a string.

        Parameters
        ----------
        conditions
            A list of SQL conditions
        """
        from .query.attribute import AttributeQuery

        self.conditions = self._match_conditions(conditions)

        self.is_fit_only = any(map(
            lambda condition: isinstance(condition, AttributeQuery),
            self.conditions
        ))

    @property
    def fit_query(self):
        subqueries = [
            f"id IN ({condition.fit_query})"
            for condition
            in self.conditions
        ]
        condition_string = f" {self.join} ".join(
            subqueries
        )
        return f"SELECT id FROM fit WHERE {condition_string}"

    @classmethod
    def _match_conditions(
            cls,
            conditions: Iterable[AbstractCondition]
    ):
        """
        Simplifies the query by matching named queries and combining junctions.

        See __init__
        """
        from .query import NamedQuery

        new_conditions = set()

        named_query_dict = defaultdict(set)

        def add_conditions(conditions_):
            for condition in conditions_:
                if condition is None:
                    continue
                if isinstance(
                        condition,
                        cls
                ):
                    add_conditions(condition)
                elif isinstance(
                        condition,
                        NamedQuery
                ) and none_table not in condition.tables:
                    named_query_dict[
                        condition.name
                    ].add(
                        condition
                    )
                else:
                    new_conditions.add(condition)

        add_conditions(conditions)

        for name, queries in named_query_dict.items():
            # noinspection PyTypeChecker
            new_conditions.add(
                NamedQuery(
                    name,
                    cls(
                        *[
                            query.other_condition
                            for query
                            in queries
                        ]
                    )
                )
            )
        return new_conditions

    def __iter__(self):
        return iter(sorted(self.conditions))

    @property
    def tables(self) -> Set[Table]:
        """
        Combines the tables of all subqueries which are not
        named queries
        """
        from .query import NamedQuery
        return {
            table
            for condition
            in self.conditions
            if not isinstance(
                condition,
                NamedQuery
            )
            for table
            in condition.tables
        }

    @property
    @abstractmethod
    def join(self) -> str:
        """
        SQL string used to conjoin queries
        """

    def __str__(self) -> str:
        """
        SQL string expressing combined query
        """
        if self.is_fit_only:
            return self.fit_query
        string = f" {self.join} ".join(map(
            str,
            sorted(
                self.conditions
            )
        ))
        return f"({string})"


class And(AbstractJunction):
    @property
    def join(self) -> str:
        return "AND"


class Or(AbstractJunction):
    @property
    def join(self):
        return "OR"



from abc import abstractmethod
from typing import Dict, Tuple, Optional, List

import numpy as np
from scipy.linalg import cho_factor

from autoconf import cached_property
from autofit.graphical.factor_graphs.abstract import AbstractNode, Value, FactorValue
# from ...mapper.operator import
from autofit.mapper.operator import (
    CholeskyOperator,
    InvCholeskyTransform,
    IdentityOperator,
    DiagonalMatrix,
)
from autofit.mapper.variable import Variable, VariableData


class VariableTransform:
    """ """

    def __init__(self, transforms):
        self.transforms = transforms

    def __mul__(self, values: Value) -> Value:
        return {k: M * values[k] for k, M in self.transforms.items()}

    def __rtruediv__(self, values: Value) -> Value:
        return {k: values[k] / M for k, M in self.transforms.items()}

    def __rmul__(self, values: Value) -> Value:
        return {k: values[k] * M for k, M in self.transforms.items()}

    def ldiv(self, values: Value) -> Value:
        return {k: M.ldiv(values[k]) for k, M in self.transforms.items()}

    def __add__(self, other: "VariableTransform") -> "VariableTransform":
        return VariableTransform(
            {k: M + other.transforms[k] for k, M in self.transforms.items()}
        )

    def inv(self) -> "VariableTransform":
        return VariableTransform({k: M.inv() for k, M in self.transforms.items()})

    rdiv = __rtruediv__
    rmul = __rmul__
    lmul = __mul__
    __matmul__ = __mul__

    def quad(self, values):
        return {v: H.T if np.ndim(H) else H for v, H in (values * self).items()} * self

    def invquad(self, values):
        return {v: H.T if np.ndim(H) else H for v, H in (values / self).items()} / self

    @cached_property
    def log_det(self):
        return sum(M.log_det for M in self.transforms.values())

    @classmethod
    def from_scales(cls, scales):
        return cls({v: DiagonalMatrix(scale) for v, scale in scales.items()})

    @classmethod
    def from_covariances(cls, covs):
        return cls(
            {v: InvCholeskyTransform(cho_factor(cov)) for v, cov in covs.items()}
        )

    @classmethod
    def from_inv_covariances(cls, inv_covs):
        return cls(
            {
                v: CholeskyOperator(cho_factor(inv_cov))
                for v, inv_cov in inv_covs.items()
            }
        )


class FullCholeskyTransform(VariableTransform):
    def __init__(self, cholesky, param_shapes):
        self.cholesky = cholesky
        self.param_shapes = param_shapes

    @classmethod
    def from_optresult(cls, opt_result):
        param_shapes = opt_result.param_shapes

        cov = opt_result.result.hess_inv
        if not isinstance(cov, np.ndarray):
            # if optimiser is L-BFGS-B then convert
            # implicit hess_inv into dense matrix
            cov = cov.todense()

        return cls(InvCholeskyTransform.from_dense(cov), param_shapes)

    def __mul__(self, values: Value) -> Value:
        M, x = self.cholesky, self.param_shapes.flatten(values)
        return self.param_shapes.unflatten(M * x)

    def __rtruediv__(self, values: Value) -> Value:
        M, x = self.cholesky, self.param_shapes.flatten(values)
        return self.param_shapes.unflatten(x / M)

    def __rmul__(self, values: Value) -> Value:
        M, x = self.cholesky, self.param_shapes.flatten(values)
        return self.param_shapes.unflatten(x * M)

    @abstractmethod
    def ldiv(self, values: Value) -> Value:
        M, x = self.cholesky, self.param_shapes.flatten(values)
        return self.param_shapes.unflatten(M.ldiv(x))

    rdiv = __rtruediv__
    rmul = __rmul__
    lmul = __mul__
    __matmul__ = __mul__

    @cached_property
    def log_det(self):
        return self.cholesky.log_det


class IdentityVariableTransform(VariableTransform):
    def __init__(self):
        pass

    def _identity(self, values: Value) -> Value:
        return values

    __mul__ = _identity
    __rtruediv__ = _identity
    __rmul__ = _identity
    ldiv = _identity
    rdiv = __rtruediv__
    rmul = __rmul__
    lmul = __mul__
    __matmul__ = __mul__
    quad = _identity
    invquad = _identity

    @property
    def log_det(self):
        return 0.0


identity_transform = IdentityOperator()
identity_variable_transform = IdentityVariableTransform()


class TransformedNode(AbstractNode):
    def __init__(self, node: AbstractNode, transform: VariableTransform):
        self.node = node
        self.transform = transform

    @property
    def variables(self):
        return self.node.variables

    @property
    def deterministic_variables(self):
        return self.node.deterministic_variables

    @property
    def all_variables(self):
        return self.node.all_variables

    @property
    def name(self):
        return f"FactorApproximation({self.node.name})"

    def __call__(
            self,
            values: Dict[Variable, np.ndarray],
    ) -> FactorValue:
        return self.node(self.transform.ldiv(values))

    def func_jacobian(
            self,
            values: Dict[Variable, np.ndarray],
            variables: Optional[List[Variable]] = None,
            _calc_deterministic: bool = True,
            **kwargs,
    ) -> Tuple[FactorValue, VariableData]:
        fval, jval = self.node.func_jacobian(
            self.transform.ldiv(values),
            variables=variables,
            _calc_deterministic=_calc_deterministic,
        )

        # TODO this doesn't deal with deterministic jacobians
        grad = jval / self.transform
        return fval, grad

    def func_jacobian_hessian(
            self,
            values: Dict[Variable, np.ndarray],
            variables: Optional[List[Variable]] = None,
            _calc_deterministic: bool = True,
            **kwargs,
    ) -> Tuple[FactorValue, VariableData, VariableData]:
        M = self.transform
        fval, jval, hval = self.node.func_jacobian_hessian(
            M.ldiv(values),
            variables=variables,
            _calc_deterministic=_calc_deterministic,
        )

        grad = jval / M
        # hess = {v: H.T for v, H in (hval / M).items()} / M
        hess = M.invquad(hval)
        return fval, grad, hess

    def __getattribute__(self, name):
        try:
            return super().__getattribute__(name)
        except AttributeError:
            return getattr(self.node, name)



from typing import List, Optional

from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.non_linear.samples.sample import Sample
from .pdf import PDFSamples


class StoredSamples(PDFSamples):

    def __init__(
            self,
            model: AbstractPriorModel,
            sample_list: List[Sample],
            unconverged_sample_size: int = 100,
            time: Optional[float] = None,
    ):
        """
        The `Samples` of a non-linear search, specifically the samples of a `NonLinearSearch` which maps out the
        posterior of parameter space and thus does provide information on parameter errors.

        Parameters
        ----------
        model : af.ModelMapper
            Maps input vectors of unit parameter values to physical values and model instances via priors.
        """

        super().__init__(
            model=model,
            sample_list=sample_list,
            time=time,
        )

        self._unconverged_sample_size = int(unconverged_sample_size)



import pickle
from functools import wraps
from typing import List

from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.non_linear.samples import Samples
from .model import Base, Object
from ..sqlalchemy_ import sa


class Pickle(Base):
    """
    A pickled python object that was found in the pickles directory
    """

    __tablename__ = "pickle"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    id = sa.Column(
        sa.Integer,
        primary_key=True
    )

    name = sa.Column(
        sa.String
    )
    string = sa.Column(
        sa.String
    )
    fit_id = sa.Column(
        sa.String,
        sa.ForeignKey(
            "fit.id"
        )
    )
    fit = sa.orm.relationship(
        "Fit",
        uselist=False
    )

    @property
    def value(self):
        """
        The unpickled object
        """
        if isinstance(
                self.string,
                str
        ):
            return self.string
        return pickle.loads(
            self.string
        )

    @value.setter
    def value(self, value):
        try:
            self.string = pickle.dumps(
                value
            )
        except pickle.PicklingError:
            pass


class Info(Base):
    __tablename__ = "info"

    id = sa.Column(
        sa.Integer,
        primary_key=True
    )

    key = sa.Column(sa.String)
    value = sa.Column(sa.String)

    fit_id = sa.Column(
        sa.String,
        sa.ForeignKey(
            "fit.id"
        )
    )
    fit = sa.orm.relationship(
        "Fit",
        uselist=False
    )


def try_none(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except TypeError:
            return None

    return wrapper


class NamedInstance(Base):
    __tablename__ = "named_instance"

    id = sa.Column(
        sa.Integer,
        primary_key=True
    )
    name = sa.Column(sa.String)

    instance_id = sa.Column(
        sa.Integer,
        sa.ForeignKey(
            "object.id"
        )
    )

    __instance = sa.orm.relationship(
        "Object",
        uselist=False,
        backref="named_instance",
        foreign_keys=[instance_id]
    )

    @property
    @try_none
    def instance(self):
        """
        An instance of the model labelled with a given name
        """
        return self.__instance()

    @instance.setter
    def instance(self, instance):
        self.__instance = Object.from_object(
            instance
        )

    fit_id = sa.Column(
        sa.String,
        sa.ForeignKey(
            "fit.id"
        )
    )
    fit = sa.orm.relationship(
        "Fit",
        uselist=False
    )


# noinspection PyProtectedMember
class NamedInstancesWrapper:
    def __init__(self, fit: "Fit"):
        """
        Provides dictionary like interface for accessing
        instance objects

        Parameters
        ----------
        fit
            A fit from which instances are accessed
        """
        self.fit = fit

    def __getitem__(self, item: str):
        """
        Get an instance with a given name.

        Raises a KeyError if no such instance exists.
        """
        return self._get_named_instance(
            item
        ).instance

    def __setitem__(self, key: str, value):
        """
        Set an instance for a given name
        """
        try:
            named_instance = self._get_named_instance(
                key
            )
        except KeyError:
            named_instance = NamedInstance(
                name=key
            )
            self.fit._named_instances.append(
                named_instance
            )
        named_instance.instance = value

    def _get_named_instance(
            self,
            item: str
    ) -> "NamedInstance":
        """
        Retrieve a NamedInstance by its name.
        """
        for named_instance in self.fit._named_instances:
            if named_instance.name == item:
                return named_instance
        raise KeyError(
            f"Instance {item} not found"
        )


class Fit(Base):
    __tablename__ = "fit"

    id = sa.Column(
        sa.String,
        primary_key=True,
    )
    is_complete = sa.Column(
        sa.Boolean
    )

    _named_instances: List[NamedInstance] = sa.orm.relationship(
        "NamedInstance"
    )

    @property
    @try_none
    def instance(self):
        """
        The instance of the model that had the highest likelihood
        """
        return self.__instance()

    @instance.setter
    def instance(self, instance):
        self.__instance = Object.from_object(
            instance
        )

    @property
    def named_instances(self):
        return NamedInstancesWrapper(
            self
        )

    _info: List[Info] = sa.orm.relationship(
        "Info"
    )

    def __init__(
            self,
            **kwargs
    ):
        super().__init__(
            **kwargs
        )

    max_log_likelihood = sa.Column(
        sa.Float
    )

    parent_id = sa.Column(
        sa.String,
        sa.ForeignKey(
            "fit.id"
        )
    )

    children: List["Fit"] = sa.orm.relationship(
        "Fit",
        backref=sa.orm.backref(
            'parent',
            remote_side=[id]
        )
    )

    @property
    def best_fit(self) -> "Fit":
        """
        Only for grid searches. Returns the child search with
        the highest log likelihood.
        """
        if not self.is_grid_search:
            raise TypeError(
                f"Fit {self.id} is not a grid search"
            )
        if len(self.children) == 0:
            raise TypeError(
                f"Grid search fit {self.id} has no children"
            )

        best_fit = None
        max_log_likelihood = float("-inf")

        for fit in self.children:
            if fit.max_log_likelihood > max_log_likelihood:
                best_fit = fit
                max_log_likelihood = fit.max_log_likelihood

        return best_fit

    is_grid_search = sa.Column(
        sa.Boolean
    )

    unique_tag = sa.Column(
        sa.String
    )
    name = sa.Column(
        sa.String
    )
    path_prefix = sa.Column(
        sa.String
    )

    _samples = sa.orm.relationship(
        Object,
        uselist=False,
        foreign_keys=[
            Object.samples_for_id
        ]
    )

    @property
    @try_none
    def samples(self) -> Samples:
        return self._samples()

    @samples.setter
    def samples(self, samples):
        self._samples = Object.from_object(
            samples
        )

    @property
    def info(self):
        return {
            info.key: info.value
            for info
            in self._info
        }

    @info.setter
    def info(self, info):
        if info is not None:
            self._info = [
                Info(
                    key=key,
                    value=value
                )
                for key, value
                in info.items()
            ]

    @property
    @try_none
    def model(self) -> AbstractPriorModel:
        """
        The model that was fit
        """
        return self.__model()

    @model.setter
    def model(self, model: AbstractPriorModel):
        self.__model = Object.from_object(
            model
        )

    pickles: List[Pickle] = sa.orm.relationship(
        "Pickle",
        lazy="joined"
    )

    def __getitem__(self, item: str):
        """
        Retrieve an object that was a pickle

        Parameters
        ----------
        item
            The name of the pickle.

            e.g. if the file were 'samples.pickle' then 'samples' would
            retrieve the unpickled object.

        Returns
        -------
        An unpickled object
        """
        for p in self.pickles:
            if p.name == item:
                return p.value
        return getattr(
            self,
            item
        )

    def __contains__(self, item):
        for p in self.pickles:
            if p.name == item:
                return True
        return False

    def __setitem__(
            self,
            key: str,
            value
    ):
        """
        Add a pickle.

        If a deserialised object is given then it is serialised
        before being added to the database.

        Parameters
        ----------
        key
            The name of the pickle
        value
            A string, bytes or object
        """
        new = Pickle(
            name=key
        )
        if isinstance(
                value,
                (str, bytes)
        ):
            new.string = value
        else:
            new.value = value
        self.pickles = [
                           p
                           for p
                           in self.pickles
                           if p.name != key
                       ] + [
                           new
                       ]

    def __delitem__(self, key):
        self.pickles = [
            p
            for p
            in self.pickles
            if p.name != key
        ]

    def value(self, name: str):
        try:
            return self.__getitem__(item=name)
        except AttributeError:
            return None

    model_id = sa.Column(
        sa.Integer,
        sa.ForeignKey(
            "object.id"
        )
    )
    __model = sa.orm.relationship(
        "Object",
        uselist=False,
        backref="fit_model",
        foreign_keys=[model_id]
    )

    instance_id = sa.Column(
        sa.Integer,
        sa.ForeignKey(
            "object.id"
        )
    )

    __instance = sa.orm.relationship(
        "Object",
        uselist=False,
        backref="fit_instance",
        foreign_keys=[instance_id]
    )

    @classmethod
    def all(cls, session):
        return session.query(
            cls
        ).all()

    def __str__(self):
        return self.id

    def __repr__(self):
        return f"<{self.__class__.__name__} {self}>"


fit_attributes = sa.inspect(Fit).columns



import logging
import numpy as np
from typing import List, Optional
import zeus

from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.non_linear.samples.mcmc import MCMCSamples
from autofit.non_linear.mcmc.auto_correlations import AutoCorrelationsSettings, AutoCorrelations
from autofit.non_linear.samples import Sample

logger = logging.getLogger(
    __name__
)

class ZeusSamples(MCMCSamples):

    @classmethod
    def from_results_internal(
            cls,
            results_internal,
            model: AbstractPriorModel,
            auto_correlation_settings: AutoCorrelationsSettings,
            unconverged_sample_size: int = 100,
            time: Optional[float] = None,
    ):
        """
        The `Samples` classes in **PyAutoFit** provide an interface between the results of a `NonLinearSearch` (e.g.
        as files on your hard-disk) and Python.

        To create a `Samples` object after an `Zeus` model-fit the results must be converted from the
        native format used by `Zeus` (which is a HDFBackend) to lists of values, the format used by the **PyAutoFit**
        `Samples` objects.

        This classmethod performs this conversion before creating a `ZeusSamples` object.

        Parameters
        ----------
        results_internal
            The MCMC results in their native internal format from which the samples are computed.
        model
            Maps input vectors of unit parameter values to physical values and model instances via priors.
        auto_correlations_settings
            Customizes and performs auto correlation calculations performed during and after the search.
        unconverged_sample_size
            If the samples are for a search that is yet to convergence, a reduced set of samples are used to provide
            a rough estimate of the parameters. The number of samples is set by this parameter.
        time
            The time taken to perform the model-fit, which is passed around `Samples` objects for outputting
            information on the overall fit.
        """

        parameter_lists = results_internal.get_chain(flat=True).tolist()
        log_posterior_list = results_internal.get_log_prob(flat=True).tolist()
        log_prior_list = model.log_prior_list_from(parameter_lists=parameter_lists)

        log_likelihood_list = [
            log_posterior - log_prior
            for log_posterior, log_prior
            in zip(log_posterior_list, log_prior_list)
        ]

        weight_list = len(log_likelihood_list) * [1.0]

        sample_list = Sample.from_lists(
            model=model,
            parameter_lists=parameter_lists,
            log_likelihood_list=log_likelihood_list,
            log_prior_list=log_prior_list,
            weight_list=weight_list
        )

        return ZeusSamples(
            model=model,
            sample_list=sample_list,
            auto_correlation_settings=auto_correlation_settings,
            unconverged_sample_size=unconverged_sample_size,
            time=time,
            results_internal=results_internal,
        )

    @property
    def samples_after_burn_in(self) -> [List]:
        """
        The zeus samples with the initial burn-in samples removed.

        The burn-in period is estimated using the auto-correlation times of the parameters.
        """

        discard = int(3.0 * np.max(self.auto_correlations.times))
        thin = int(np.max(self.auto_correlations.times) / 2.0)
        return self.results_internal.get_chain(discard=discard, thin=thin, flat=True)

    @property
    def total_walkers(self):
        return len(self.results_internal.get_chain()[0, :, 0])

    @property
    def total_steps(self):
        return int(self.results_internal.ncall_total)

    @property
    def auto_correlations(self):

        times = zeus.AutoCorrTime(samples=self.results_internal.get_chain())
        try:
            previous_auto_correlation_times = zeus.AutoCorrTime(
                samples=self.results_internal.get_chain()[: -self.auto_correlation_settings.check_size, :, :],
            )
        except IndexError:
            logger.debug(
                "Unable to compute previous auto correlation times."
            )
            previous_auto_correlation_times = None

        return AutoCorrelations(
            check_size=self.auto_correlation_settings.check_size,
            required_length=self.auto_correlation_settings.required_length,
            change_threshold=self.auto_correlation_settings.change_threshold,
            times=times,
            previous_times=previous_auto_correlation_times,
        )



import matplotlib.pyplot as plt
from autofit.plot.output import Output

class SamplesPlotter:
    def __init__(
            self, 
            samples,
            output : Output = Output()
    ):

        self.samples = samples
        self.output = output

    @property
    def model(self):
        return self.samples.model

    def close(self):
        if plt.fignum_exists(num=1):
            plt.close()

class MCMCPlotter(SamplesPlotter):

    def _plot_trajectories(self, samples, log_posterior_list, **kwargs):

        fig, axes = plt.subplots(self.samples.model.prior_count, figsize=(10, 7))

        for i in range(self.samples.model.prior_count):

            for walker_index in range(log_posterior_list.shape[1]):

                ax = axes[i]
                ax.plot(samples[:, walker_index, i], log_posterior_list[:, walker_index], alpha=0.3)

            ax.set_ylabel("Log Likelihood")
            ax.set_xlabel(self.model.parameter_labels_with_superscripts_latex[i])

        self.output.to_figure(structure=None, auto_filename="tracjectories")
        self.close()

    def _plot_likelihood_series(self, log_posterior_list, **kwargs):

        fig, axes = plt.subplots(1, figsize=(10, 7))

        for walker_index in range(log_posterior_list.shape[1]):

            axes.plot(log_posterior_list[:, walker_index], alpha=0.3)

        axes.set_ylabel("Log Likelihood")
        axes.set_xlabel("step number")

        self.output.to_figure(structure=None, auto_filename="likelihood_series")
        self.close()

    def _plot_time_series(self, samples, **kwargs):

        fig, axes = plt.subplots(self.samples.model.prior_count, figsize=(10, 7), sharex=True)

        for i in range(self.samples.model.prior_count):
            ax = axes[i]
            ax.plot(samples[:, :, i], alpha=0.3)
            ax.set_ylabel(self.model.parameter_labels_with_superscripts_latex[i])

        axes[-1].set_xlabel("step number")

        self.output.to_figure(structure=None, auto_filename="time_series")
        self.close()


import logging

from autofit.text import formatter as frm

logger = logging.getLogger(__name__)


def values_from_samples(samples, median_pdf_model):

    if median_pdf_model:
        return samples.median_pdf_vector
    return samples.max_log_likelihood_vector


def summary(
    samples, sigma=3.0, median_pdf_model=True, indent=1, line_length=None
) -> str:
    """ Create a string summarizing the results of the `NonLinearSearch` at an input sigma value.

    This function is used for creating the model.results files of a non-linear search.

    Parameters
    ----------
    sigma
        The sigma within which the PDF is used to estimate errors (e.g. sigma = 1.0 uses 0.6826 of the PDF)."""

    values = values_from_samples(samples=samples, median_pdf_model=median_pdf_model)
    values_at_sigma = samples.vector_at_sigma(sigma=sigma)

    parameter_names = samples.model.parameter_names

    if line_length is None:
        line_length = len(max(parameter_names, key=len)) + 8

    sigma_formatter = frm.TextFormatter(indent=indent, line_length=line_length)

    for i, prior_path in enumerate(samples.model.unique_prior_paths):

        value_result = frm.value_result_string_from(
            parameter_name=parameter_names[i],
            value=values[i],
            values_at_sigma=values_at_sigma[i],
        )

        sigma_formatter.add(prior_path, value_result)

    return f"\n\nSummary ({sigma} sigma limits):\n\n{sigma_formatter.text}"


def latex(
        samples,
        median_pdf_model=True,
        sigma=3.0,
        name_to_label=True,
        include_name=True,
        include_quickmath=False,
        prefix="",
        suffix=""
) -> str:
    """
    Create a string summarizing the results of the `NonLinearSearch` at an input sigma value.

    This function is used for creating the model.results files of a non-linear search.

    Parameters
    ----------
    sigma
        The sigma within which the PDF is used to estimate errors (e.g. sigma = 1.0 uses 0.6826 of the PDF)."""

    values = values_from_samples(samples=samples, median_pdf_model=median_pdf_model)
    errors_at_sigma = samples.error_vector_at_sigma(sigma=sigma)

    table = []

    for i in range(samples.model.prior_count):

        label_value = frm.parameter_result_latex_from(
            parameter_name=samples.model.parameter_names[i],
            value=values[i],
            errors=errors_at_sigma[i],
            superscript=samples.model.superscripts[i],
            name_to_label=name_to_label,
            include_name=include_name,
            include_quickmath=include_quickmath
        )

        table.append(f"{label_value}")

    table = "".join(table)[:-3]

    return f"{prefix}{table}{suffix}"



from autofit.messages.normal import NormalMessage
from .abstract import Prior


class GaussianPrior(Prior):
    """A prior with a gaussian distribution"""

    __identifier_fields__ = (
        "lower_limit",
        "upper_limit",
        "mean",
        "sigma"
    )

    def __init__(
            self,
            mean,
            sigma,
            lower_limit=float("-inf"),
            upper_limit=float("inf"),

    ):
        super().__init__(
            lower_limit=lower_limit,
            upper_limit=upper_limit,
            message=NormalMessage(
                mean=mean,
                sigma=sigma,
                lower_limit=lower_limit,
                upper_limit=upper_limit,
            )
        )

    @classmethod
    def with_limits(
            cls,
            lower_limit: float,
            upper_limit: float
    ) -> "GaussianPrior":
        """
        Create a new gaussian prior centred between two limits
        with sigma distance between this limits.

        Note that these limits are not strict so exceptions will not
        be raised for values outside of the limits.

        Parameters
        ----------
        lower_limit
        upper_limit

        Returns
        -------
        A new GaussianPrior
        """
        return cls(
            mean=(lower_limit + upper_limit) / 2,
            sigma=upper_limit - lower_limit,
        )

    def dict(self) -> dict:
        """
        A dictionary representation of this prior
        """
        prior_dict = super().dict()
        return {**prior_dict, "mean": self.mean, "sigma": self.sigma}

    def __str__(self):
        """The line of text describing this prior for the model_mapper.info file"""
        return f"GaussianPrior, mean = {self.mean}, sigma = {self.sigma}"



from os import path
from typing import Optional

from autofit.database.sqlalchemy_ import sa

from autoconf import conf
from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.non_linear.optimize.abstract_optimize import AbstractOptimizer
from autofit.non_linear.abstract_search import PriorPasser
from autofit.non_linear.initializer import AbstractInitializer
from autofit.non_linear.optimize.drawer.plotter import DrawerPlotter
from autofit.non_linear.samples import Samples, Sample
from autofit.plot.output import Output


class Drawer(AbstractOptimizer):

    __identifier_fields__ = (
        "total_draws",
    )

    def __init__(
            self,
            name: Optional[str] = None,
            path_prefix: Optional[str] = None,
            unique_tag: Optional[str] = None,
            prior_passer: Optional[PriorPasser] = None,
            initializer: Optional[AbstractInitializer] = None,
            iterations_per_update: int = None,
            session: Optional[sa.orm.Session] = None,
            **kwargs
    ):
        """
        A Drawer non-linear search, which simply draws a fixed number of samples from the model uniformly from the
        priors.

        Therefore, it does not seek to determine model parameters which maximize the likelihood or map out the
        posterior of the overall parameter space.

        Whilst this is not the typical use case of a non-linear search, it has certain niche applications, for example:

        - Given a model one can determine how much variation there is in the log likelihood / log posterior values.
        By visualizing this as a histogram one can therefore quantify the behaviour of that
        model's `log_likelihood_function`.

        - If the `log_likelihood_function` of a model is stochastic (e.g. different values of likelihood may be
        computed for an identical model due to randomness in the likelihood evaluation) this search can quantify
        the behaviour of that stochasticity.

        - For advanced modeling tools, for example sensitivity mapping performed via the `Sensitivity` object,
        the `Drawer` search may be sufficient to perform the overall modeling task, without the need of performing
        an actual parameter space search.

        The drawer search itself is performed by simply reusing the functionality of the `AbstractInitializer` object.
        Whereas this is normally used to initialize a non-linear search, for the drawer it performed all log
        likelihood evluations.

        Parameters
        ----------
        name
            The name of the search, controlling the last folder results are output.
        path_prefix
            The path of folders prefixing the name folder where results are output.
        unique_tag
            The name of a unique tag for this model-fit, which will be given a unique entry in the sqlite database
            and also acts as the folder after the path prefix and before the search name.
        prior_passer
            Controls how priors are passed from the results of this `NonLinearSearch` to a subsequent non-linear search.
        initializer
            Generates the initialize samples of non-linear parameter space (see autofit.non_linear.initializer).
        session
            An SQLalchemy session instance so the results of the model-fit are written to an SQLite database.
        """

        super().__init__(
            name=name,
            path_prefix=path_prefix,
            unique_tag=unique_tag,
            prior_passer=prior_passer,
            initializer=initializer,
            iterations_per_update=iterations_per_update,
            session=session,
            **kwargs
        )

        self.number_of_cores = 1

        self.logger.debug("Creating Drawer Search")

    class Fitness(AbstractOptimizer.Fitness):

        def figure_of_merit_from(self, parameter_list):
            """
            The figure of merit is the value that the `NonLinearSearch` uses to sample parameter space.

            The `Drawer` search can use either the log posterior values or log likelihood values.
            """
            return self.log_posterior_from(parameter_list=parameter_list)

    def _fit(self, model: AbstractPriorModel, analysis, log_likelihood_cap=None):
        """
        Fit a model using Drawer and the Analysis class which contains the data and returns the log likelihood from
        instances of the model, which the `NonLinearSearch` seeks to maximize.

        Parameters
        ----------
        model : ModelMapper
            The model which generates instances for different points in parameter space.
        analysis : Analysis
            Contains the data and the log likelihood function which fits an instance of the model to the data, returning
            the log likelihood the `NonLinearSearch` maximizes.

        Returns
        -------
        A result object comprising the Samples object that inclues the maximum log likelihood instance and full
        chains used by the fit.
        """

        fitness_function = self.fitness_function_from_model_and_analysis(
            model=model, analysis=analysis
        )

        total_draws = self.config_dict_search["total_draws"]

        self.logger.info(f"Performing DrawerSearch for a total of {total_draws} points.")

        unit_parameter_lists, parameter_lists, log_posterior_list = self.initializer.samples_from_model(
            total_points=self.config_dict_search["total_draws"],
            model=model,
            fitness_function=fitness_function,
            use_prior_medians=True
        )

        self.paths.save_object(
            "parameter_lists",
            parameter_lists
        )
        self.paths.save_object(
            "log_posterior_list",
            log_posterior_list
        )

        self.perform_update(
            model=model, analysis=analysis, during_analysis=False
        )

        self.logger.info("Drawer complete")

    def fitness_function_from_model_and_analysis(self, model, analysis, log_likelihood_cap=None):

        return Drawer.Fitness(
            paths=self.paths,
            model=model,
            analysis=analysis,
            samples_from_model=self.samples_from,
            log_likelihood_cap=log_likelihood_cap,
        )

    def samples_from(self, model):

        parameter_lists = self.paths.load_object("parameter_lists")

        log_prior_list = [
            sum(model.log_prior_list_from_vector(vector=vector))
            for vector
            in parameter_lists
        ]
        log_likelihood_list = [
            lp - prior for lp, prior
            in zip(self.paths.load_object("log_posterior_list"), log_prior_list)
        ]

        weight_list = len(log_likelihood_list) * [1.0]

        sample_list = Sample.from_lists(
            model=model,
            parameter_lists=parameter_lists,
            log_likelihood_list=log_likelihood_list,
            log_prior_list=log_prior_list,
            weight_list=weight_list
        )

        return Samples(
            model=model,
            sample_list=sample_list,
            time=self.timer.time
        )

    def plot_results(self, samples):

        def should_plot(name):
            return conf.instance["visualize"]["plots_search"]["drawer"][name]

        plotter = DrawerPlotter(
            samples=samples,
            output=Output(path=path.join(self.paths.image_path, "search"), format="png")
        )




from autofit.messages.normal import UniformNormalMessage
from .abstract import Prior
from .abstract import epsilon
from .wrapped_instance import WrappedInstance


class UniformPrior(Prior):
    """A prior with a uniform distribution between a lower and upper limit"""

    def __init__(
            self,
            lower_limit=0.0,
            upper_limit=1.0,
            id_=None,
    ):
        lower_limit = float(lower_limit)
        upper_limit = float(upper_limit)

        Message = UniformNormalMessage.shifted(
            shift=lower_limit,
            scale=upper_limit - lower_limit,
        )
        super().__init__(
            WrappedInstance(
                Message,
                0, 1,
                lower_limit=lower_limit,
                upper_limit=upper_limit,
                id_=id_,
            ),
            lower_limit=lower_limit,
            upper_limit=upper_limit,
            id_=id_,
        )

    @property
    def lower_unit_limit(self) -> float:
        """
        The lower limit for this prior in unit vector space
        """
        return 0.0

    @property
    def upper_unit_limit(self) -> float:
        """
        The upper limit for this prior in unit vector space
        """
        return 1.0

    def logpdf(self, x):
        # TODO: handle x as a numpy array
        if x == self.lower_limit:
            x += epsilon
        elif x == self.upper_limit:
            x -= epsilon
        return self.instance().logpdf(x)

    def __str__(self):
        """The line of text describing this prior for the model_mapper.info file"""
        return f"UniformPrior, lower_limit = {self.lower_limit}, upper_limit = {self.upper_limit}"

    def value_for(self, unit, ignore_prior_limits=False):
        """

        Parameters
        ----------
        unit: Float
            A unit hypercube value between 0 and 1
        Returns
        -------
        value: Float
            A value for the attribute between the upper and lower limits
        """
        return round(super().value_for(unit, ignore_prior_limits=ignore_prior_limits), 14)

    # noinspection PyUnusedLocal
    @staticmethod
    def log_prior_from_value(value):
        """
        Returns the log prior of a physical value, so the log likelihood of a model evaluation can be converted to a
        posterior as log_prior + log_likelihood.

        This is used by Emcee in the log likelihood function evaluation.

        NOTE: For a UniformPrior this is always zero, provided the value is between the lower and upper limit. Given
        this is check for when the instance is made (in the *instance_from_vector* function), we thus can simply return
        zero in this function.
        """
        return 0.0



import logging
from typing import Tuple

from autofit.mapper.prior.abstract import Prior
from autofit.mapper.prior.tuple_prior import TuplePrior
from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.mapper.prior_model.collection import CollectionPriorModel
from .analysis import Analysis
from .indexed import IndexCollectionAnalysis

logger = logging.getLogger(
    __name__
)


class FreeParameterAnalysis(IndexCollectionAnalysis):
    def __init__(
            self,
            *analyses: Analysis,
            free_parameters: Tuple[Prior, ...]
    ):
        """
        A combined analysis with free parameters.

        All parameters for the model are shared across every analysis except
        for the free parameters which are allowed to vary for individual
        analyses.

        Parameters
        ----------
        analyses
            A list of analyses
        free_parameters
            A list of priors which are independent for each analysis
        """
        super().__init__(*analyses)
        self.free_parameters = [
            parameter for parameter
            in free_parameters
            if isinstance(
                parameter,
                Prior
            )
        ]
        # noinspection PyUnresolvedReferences
        self.free_parameters += [
            prior
            for parameter
            in free_parameters
            if isinstance(
                parameter,
                (AbstractPriorModel, TuplePrior)
            )
            for prior
            in parameter.priors
        ]

    def modify_model(
            self,
            model: AbstractPriorModel
    ) -> AbstractPriorModel:
        """
        Create prior models where free parameters are replaced with new
        priors. Return those prior models as a collection.

        The number of dimensions of the new prior model is the number of the
        old one plus the number of free parameters multiplied by the number
        of free parameters.

        Parameters
        ----------
        model
            The original model

        Returns
        -------
        A new model with all the same priors except for those associated
        with free parameters.
        """
        return CollectionPriorModel([
            model.mapper_from_partial_prior_arguments({
                free_parameter: free_parameter.new()
                for free_parameter in self.free_parameters
            })
            for _ in self.analyses
        ])



from typing import Union

from autofit.graphical.declarative.factor.hierarchical import HierarchicalFactor
from autofit.mapper.model import ModelInstance
from autofit.tools.namer import namer
from .abstract import AbstractDeclarativeFactor


class FactorGraphModel(AbstractDeclarativeFactor):
    def __init__(
            self,
            *model_factors: Union[
                AbstractDeclarativeFactor,
                HierarchicalFactor
            ],
            name=None,
            include_prior_factors=True,
    ):
        """
        A collection of factors that describe models, which can be
        used to create a graph and messages.

        If the models have shared priors then the graph has shared variables

        Parameters
        ----------
        model_factors
            Factors which are hierarchical or associated with a specific analysis
        """
        super().__init__(
            include_prior_factors=include_prior_factors,
        )
        self._model_factors = list(model_factors)
        self._name = name or namer(self.__class__.__name__)

    @property
    def prior_model(self):
        """
        Construct a CollectionPriorModel comprising the prior models described
        in each model factor
        """
        from autofit.mapper.prior_model.collection import CollectionPriorModel
        return CollectionPriorModel({
            factor.name: factor.prior_model
            for factor
            in self.model_factors
        })

    @property
    def optimiser(self):
        raise NotImplemented()

    @property
    def info(self) -> str:
        """
        Info describing this collection.
        """
        return self.graph.info

    @property
    def name(self):
        return self._name

    def add(
            self,
            model_factor: AbstractDeclarativeFactor
    ):
        """
        Add another factor to this collection.
        """
        self._model_factors.append(
            model_factor
        )

    def log_likelihood_function(
            self,
            instance: ModelInstance
    ) -> float:
        """
        Compute the combined likelihood of each factor from a collection of instances
        with the same ordering as the factors.

        Parameters
        ----------
        instance
            A collection of instances, one corresponding to each factor

        Returns
        -------
        The combined likelihood of all factors
        """
        log_likelihood = 0
        for model_factor, instance_ in zip(
                self.model_factors,
                instance
        ):
            log_likelihood += model_factor.log_likelihood_function(
                instance_
            )

        return log_likelihood

    @property
    def model_factors(self):
        model_factors = list()
        for model_factor in self._model_factors:
            if isinstance(
                    model_factor,
                    HierarchicalFactor
            ):
                model_factors.extend(
                    model_factor.factors
                )
            else:
                model_factors.append(
                    model_factor
                )
        return model_factors



from typing import Optional

from .abstract import AbstractPaths


class NullPaths(AbstractPaths):
    """
    Null version of paths object for avoiding writing of files to disk
    """

    def __init__(self):
        super().__init__()
        self.objects = dict()

    @AbstractPaths.parent.setter
    def parent(self, parent):
        pass

    @property
    def is_grid_search(self) -> bool:
        return False

    def create_child(
            self,
            name: Optional[str] = None,
            path_prefix: Optional[str] = None,
            is_identifier_in_paths: Optional[bool] = None,
            identifier: Optional[str] = None
    ) -> "AbstractPaths":
        return NullPaths()

    def save_named_instance(self, name: str, instance):
        pass

    def save_object(self, name: str, obj: object):
        self.objects[name] = obj

    def load_object(self, name: str):
        return self.objects[name]

    def remove_object(self, name: str):
        pass

    def is_object(self, name: str) -> bool:
        pass

    @property
    def is_complete(self) -> bool:
        return False

    def completed(self):
        pass

    def save_all(self, search_config_dict, info, pickle_files):
        pass

    def load_samples(self):
        pass

    def save_samples(self, samples):
        pass

    def load_samples_info(self):
        pass



from typing import Type, Tuple, Union

import numpy as np

from autoconf import cached_property
from .abstract import AbstractMessage
from .transform import AbstractDensityTransform


class TransformedMessage(AbstractMessage):
    _Message: Type[AbstractMessage]
    _transform: Union[AbstractDensityTransform, Type[AbstractDensityTransform]]
    _depth = 0

    def __init__(self, *args, **kwargs):
        self.instance = self._Message(*args, **kwargs)
        super().__init__(
            *self.instance.parameters,
            log_norm=self.instance.log_norm,
            lower_limit=self.instance.lower_limit,
            upper_limit=self.instance.upper_limit,
            id_=self.instance.id
        )

    def cdf(self, x):
        return self.instance.cdf(
            self._transform.transform(x)
        )

    @property
    def natural_parameters(self):
        return self.instance.natural_parameters

    @property
    def log_partition(self) -> np.ndarray:
        return self.instance.log_partition

    def __getattr__(self, item):
        if item == "__setstate__":
            raise AttributeError()
        return getattr(
            self.instance,
            item
        )

    @classmethod
    def invert_natural_parameters(
            cls,
            natural_parameters
    ):
        return cls._Message.invert_natural_parameters(
            natural_parameters
        )

    @classmethod
    def invert_sufficient_statistics(
            cls,
            sufficient_statistics
    ):
        return cls._Message.invert_sufficient_statistics(
            sufficient_statistics
        )

    def value_for(self, unit):
        return self._transform.inv_transform(
            self.instance.value_for(
                unit
            )
        )

    # noinspection PyMethodOverriding
    @classmethod
    def _reconstruct(  # type: ignore
            cls,
            Message: 'AbstractMessage',
            clsname: str,
            transform: AbstractDensityTransform,
            parameters: Tuple[np.ndarray, ...],
            log_norm: float,
            id_,
            lower_limit,
            upper_limit
    ):
        # Reconstructs TransformedMessage during unpickling
        Transformed = Message.transformed(transform, clsname)
        return Transformed(
            *parameters,
            log_norm=log_norm,
            id_=id_,
            lower_limit=lower_limit,
            upper_limit=upper_limit
        )

    def __reduce__(self):
        # serialises TransformedMessage during pickling
        return (
            TransformedMessage._reconstruct,
            (
                self._Message,
                self.__class__.__name__,
                self._transform,
                self.parameters,
                self.log_norm,
                self.id,
                self.lower_limit,
                self.upper_limit
            ),
        )

    @classmethod
    def calc_log_base_measure(cls, x) -> np.ndarray:
        x = cls._transform.transform(x)
        log_base = cls._Message.calc_log_base_measure(x)
        return log_base

    @classmethod
    def to_canonical_form(cls, x) -> np.ndarray:
        x = cls._transform.transform(x)
        return cls._Message.to_canonical_form(x)

    @cached_property
    def mean(self) -> np.ndarray:
        # noinspection PyUnresolvedReferences
        return self._transform.inv_transform(
            self.instance.mean
        )

    # @cached_property
    @property
    def variance(self) -> np.ndarray:
        # noinspection PyUnresolvedReferences
        jac = self._transform.jacobian(self.mean)
        return jac.quad(self._Message.variance.func(self))

    def _sample(self, n_samples) -> np.ndarray:
        x = self.instance._sample(n_samples)
        return self._transform.inv_transform(x)

    @classmethod
    def _factor(
            cls,
            self,
            x: np.ndarray,
    ) -> np.ndarray:
        x, log_det = cls._transform.transform_det(x)
        eta = self._broadcast_natural_parameters(x)
        t = cls._Message.to_canonical_form(x)
        log_base = self.calc_log_base_measure(x) + log_det
        return self.natural_logpdf(eta, t, log_base, self.log_partition)

    @classmethod
    def _factor_gradient(
            cls,
            self,
            x: np.ndarray,
    ) -> np.ndarray:
        x, logd, logd_grad, jac = cls._transform.transform_det_jac(x)
        logl, grad = cls._Message._logpdf_gradient(self, x)
        return logl + logd, grad * jac + logd_grad

    def factor(self, x):
        return self._factor(self, x)

    def factor_gradient(self, x):
        return self._factor_gradient(self, x)

    @classmethod
    def _logpdf_gradient(  # type: ignore
            cls,
            self,
            x: np.ndarray,
    ) -> Tuple[np.ndarray, np.ndarray]:
        x, jac = cls._transform.transform_jac(x)
        logl, grad = cls._Message._logpdf_gradient(self, x)
        return logl, grad * jac

    def sample(self, n_samples=None) -> np.ndarray:
        return self._sample(n_samples)

    def logpdf_gradient(
            self, x: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        return self._logpdf_gradient(self, x)

    # @classmethod
    # def _logpdf_gradient_hessian(  # type: ignore
    #         cls,
    #         self,
    #         x: np.ndarray,
    # ) -> Tuple[np.ndarray, np.ndarray]:
    #     x, jac = cls._transform.transform_jac(x)
    #     logl, grad, hess = cls._Message._logpdf_gradient_hessian(self, x)
    #     return logl, grad * jac, jac.quad(hess)

    # def logpdf_gradient_hessian(
    #         self, x: np.ndarray
    # ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    #     return self._logpdf_gradient_hessian(self, x)



import itertools
from typing import Type

from autoconf.class_path import get_class
from .identifier import Identifier


class ModelObject:
    _ids = itertools.count()

    @classmethod
    def next_id(cls):
        return next(cls._ids)

    def __init__(
            self,
            id_=None,
            label=None,
    ):
        """
        A generic object in AutoFit

        Parameters
        ----------
        id_
            A unique integer identifier. This is used to hash and order priors.
        label
            A label which can optionally be set for visualising this object in a
            graph.
        """
        self.id = id_ or self.next_id()
        self._label = label

    def has(self, cls: Type) -> bool:
        """
        Does this instance have an attribute which is of type cls?
        """
        for value in self.__dict__.values():
            if isinstance(value, cls):
                return True
        return False

    @property
    def label(self):
        return self._label

    @label.setter
    def label(self, label):
        self._label = label

    @property
    def component_number(self):
        return self.id

    def __hash__(self):
        return self.id

    def __eq__(self, other):
        try:
            return self.id == other.id
        except AttributeError:
            return False

    @property
    def identifier(self):
        return str(Identifier(self))

    @staticmethod
    def from_dict(d):
        """
        Recursively parse a dictionary returning the model, collection or
        instance that is represents.

        Parameters
        ----------
        d
            A dictionary representation of some object

        Returns
        -------
        An instance
        """
        from autofit.mapper.prior_model.abstract import AbstractPriorModel
        from autofit.mapper.prior_model.collection import CollectionPriorModel
        from autofit.mapper.prior_model.prior_model import PriorModel
        from autofit.mapper.prior.abstract import Prior
        from autofit.mapper.prior.tuple_prior import TuplePrior

        if not isinstance(
                d, dict
        ):
            return d

        type_ = d["type"]

        if type_ == "model":
            instance = PriorModel(
                get_class(
                    d.pop("class_path")
                )
            )
        elif type_ == "collection":
            instance = CollectionPriorModel()
        elif type_ == "instance":
            cls = get_class(
                d.pop("class_path")
            )
            instance = object.__new__(cls)
        elif type_ == "tuple_prior":
            instance = TuplePrior()
        else:
            return Prior.from_dict(d)

        d.pop("type")

        for key, value in d.items():
            setattr(
                instance,
                key,
                AbstractPriorModel.from_dict(value)
            )
        return instance

    def dict(self) -> dict:
        """
        A dictionary representation of this object
        """
        from autofit.mapper.prior_model.abstract import AbstractPriorModel
        from autofit.mapper.prior_model.collection import CollectionPriorModel
        from autofit.mapper.prior_model.prior_model import PriorModel
        from autofit.mapper.prior.tuple_prior import TuplePrior

        if isinstance(
                self,
                CollectionPriorModel
        ):
            type_ = "collection"
        elif isinstance(
                self,
                AbstractPriorModel
        ) and self.prior_count == 0:
            type_ = "instance"
        elif isinstance(
                self,
                PriorModel
        ):
            type_ = "model"
        elif isinstance(
                self,
                TuplePrior
        ):
            type_ = "tuple_prior"
        else:
            raise AssertionError(
                f"{self.__class__.__name__} cannot be serialised to dict"
            )

        dict_ = {
            "type": type_
        }

        for key, value in self._dict.items():
            try:
                if not isinstance(
                        value, ModelObject
                ):
                    value = AbstractPriorModel.from_instance(
                        value
                    )
                value = value.dict()
            except AttributeError:
                pass
            except TypeError:
                pass
            dict_[key] = value
        return dict_

    @property
    def _dict(self):
        return {
            key: value
            for key, value in self.__dict__.items()
            if key not in ("component_number", "item_number", "id", "cls")
               and not key.startswith("_")
        }



import numpy as np

from typing import Optional

class AutoCorrelationsSettings:

    def __init__(
            self,
            check_for_convergence: Optional[bool] = None,
            check_size: Optional[int] = None,
            required_length: Optional[int] = None,
            change_threshold: Optional[float] = None,
    ):
        """
        Class for performing and customizing AutoCorrelation calculations, which are used:

         - By the `Samples` object during a model-fit to determine is an ensemble MCMC sampler should terminate.
         - After a model-fit is finished to investigate whether the resutls converged.

        Parameters
        ----------
        check_for_convergence
            Whether the auto-correlation lengths of the Emcee samples are checked to determine the stopping criteria.
            If `True`, this option may terminate the Emcee run before the input number of steps, nsteps, has
            been performed. If `False` nstep samples will be taken.
        check_size
            The length of the samples used to check the auto-correlation lengths (from the latest sample backwards).
            For convergence, the auto-correlations must not change over a certain range of samples. A longer check-size
            thus requires more samples meet the auto-correlation threshold, taking longer to terminate sampling.
            However, shorter chains risk stopping sampling early due to noise.
        required_length
            The length an auto_correlation chain must be for it to be used to evaluate whether its change threshold is
            sufficiently small to terminate sampling early.
        change_threshold
            The threshold value by which if the change in auto_correlations is below sampling will be terminated early.
        """
        self.check_for_convergence = check_for_convergence
        self.check_size = check_size
        self.required_length = required_length
        self.change_threshold = change_threshold

    def update_via_config(self, config):

        config_dict = config._dict

        self.check_for_convergence = self.check_for_convergence if self.check_for_convergence is not None else config_dict["check_for_convergence"]
        self.check_size = self.check_size or config_dict["check_size"]
        self.required_length = self.required_length or config_dict["required_length"]
        self.change_threshold = self.change_threshold or config_dict["change_threshold"]


class AutoCorrelations(AutoCorrelationsSettings):

    def __init__(
            self,
            check_size,
            required_length,
            change_threshold,
            times,
            previous_times,

    ):
        """
        Class for performing and customizing AutoCorrelation calculations, which are used:

         - By the `Samples` object during a model-fit to determine is an ensemble MCMC sampler should terminate.
         - After a model-fit is finished to investigate whether the resutls converged.

        Parameters
        ----------
        check_for_convergence
            Whether the auto-correlation lengths of the Emcee samples are checked to determine the stopping criteria.
            If `True`, this option may terminate the Emcee run before the input number of steps, nsteps, has
            been performed. If `False` nstep samples will be taken.
        check_size
            The length of the samples used to check the auto-correlation lengths (from the latest sample backwards).
            For convergence, the auto-correlations must not change over a certain range of samples. A longer check-size
            thus requires more samples meet the auto-correlation threshold, taking longer to terminate sampling.
            However, shorter chains risk stopping sampling early due to noise.
        required_length
            The length an auto_correlation chain must be for it to be used to evaluate whether its change threshold is
            sufficiently small to terminate sampling early.
        change_threshold
            The threshold value by which if the change in auto_correlations is below sampling will be terminated early.
        """

        super().__init__(
            check_size=check_size,
            required_length=required_length,
            change_threshold=change_threshold
        )

        self.times = times
        self.previous_times = previous_times

    @property
    def relative_times(self) -> [float]:
        return np.abs(self.previous_times - self.times) / self.times

    def check_if_converged(self, total_samples):
        """
        Whether the emcee samples have converged on a solution or if they are still in a burn-in period, based on the
        auto correlation times of parameters.
        """

        if self.times is None or self.previous_times is None:
            return False

        converged = np.all(
            self.times * self.required_length
            < total_samples
        )
        if converged:
            try:
                converged &= np.all(self.relative_times < self.change_threshold)
            except IndexError:
                return False
        return converged



import numpy as np

from autofit import exc
from autofit.messages.normal import UniformNormalMessage
from autofit.messages.transform import log_10_transform
from .abstract import Prior
from .wrapped_instance import WrappedInstance


class LogUniformPrior(Prior):
    """A prior with a uniform distribution between a lower and upper limit"""

    def __init__(
            self,
            lower_limit=1e-6,
            upper_limit=1.0,
            id_=None,
    ):
        if lower_limit <= 0.0:
            raise exc.PriorException(
                "The lower limit of a LogUniformPrior cannot be zero or negative."
            )

        lower_limit = float(lower_limit)
        upper_limit = float(upper_limit)

        Message = UniformNormalMessage.shifted(
            shift=np.log10(lower_limit),
            scale=np.log10(upper_limit / lower_limit),
        ).transformed(
            log_10_transform
        )

        super().__init__(
            message=WrappedInstance(
                Message,
                0.0, 1.0,
                id_=id_,
                lower_limit=lower_limit,
                upper_limit=upper_limit,
            ),
            lower_limit=lower_limit,
            upper_limit=upper_limit,
            id_=id_,
        )

    @classmethod
    def with_limits(
            cls,
            lower_limit: float,
            upper_limit: float
    ):
        return cls(
            lower_limit=max(
                0.000001,
                lower_limit
            ),
            upper_limit=upper_limit,
        )

    __identifier_fields__ = ("lower_limit", "upper_limit")

    @staticmethod
    def log_prior_from_value(value):
        """
        Returns the log prior of a physical value, so the log likelihood of a model evaluation can be converted to a
            posterior as log_prior + log_likelihood.

        This is used by Emcee in the log likelihood function evaluation.

        Parameters
        ----------
        value : float
            The physical value of this prior's corresponding parameter in a `NonLinearSearch` sample."""
        return 1.0 / value

    def value_for(self, unit: float, ignore_prior_limits=False) -> float:
        return super().value_for(unit, ignore_prior_limits=ignore_prior_limits)

    def __str__(self):
        """The line of text describing this prior for the model_mapper.info file"""
        return f"LogUniformPrior, lower_limit = {self.lower_limit}, upper_limit = {self.upper_limit}"



import logging
import warnings
from typing import Optional, Dict, Tuple, Any, Callable

import numpy as np

from autofit import exc
from autofit.graphical.laplace.line_search import line_search, OptimisationState
from autofit.graphical.utils import Status, StatusFlag, LogWarnings
from autofit.mapper.variable_operator import VariableData


## get ascent direction


def gradient_ascent(state: OptimisationState, **kwargs) -> VariableData:
    return state.gradient


def newton_direction(state: OptimisationState, **kwargs) -> VariableData:
    return state.hessian.ldiv(state.gradient)

def newton_abs_direction(state: OptimisationState, d=1e-6, **kwargs) -> VariableData:
    posdef = state.hessian.abs().diagonalupdate(state.parameters.full_like(d))
    return posdef.ldiv(state.gradient)


logger = logging.getLogger(__name__)
logging.captureWarnings(False)
_log_projection_warnings = logger.debug


## Quasi-newton approximations


def sr1_update(
        state1: OptimisationState, state: OptimisationState, mintol=1e-8, **kwargs
) -> OptimisationState:
    yk = VariableData.sub(state1.gradient, state.gradient)
    dk = VariableData.sub(state1.parameters, state.parameters)
    Bk = state.hessian
    zk = yk + Bk * dk
    zkdk = -zk.dot(dk)

    tol = mintol * dk.norm() * zk.norm()
    if zkdk > tol:
        vk = zk / np.sqrt(zkdk)
        Bk1 = Bk.lowrankupdate(vk)
    elif zkdk < -tol:
        vk = zk / np.sqrt(-zkdk)
        Bk1 = Bk.lowrankdowndate(vk)
    else:
        Bk1 = Bk

    state1.hessian = Bk1
    return state1


def diag_sr1_update(
        state1: OptimisationState, state: OptimisationState, tol=1e-8, **kwargs
) -> OptimisationState:
    yk = VariableData.sub(state1.gradient, state.gradient)
    dk = VariableData.sub(state1.parameters, state.parameters)
    Bk = state.hessian
    zk = yk + Bk * dk
    dzk = dk * zk
    # alpha = -zk.dot(dk) / dzk.dot(dzk)

    d = dzk.dot(dzk)
    if d > tol * dk.norm() ** 2 * zk.norm() ** 2:
        alpha = -zk.dot(dk) / d
        Bk = Bk.diagonalupdate(alpha * (zk ** 2))

    state1.hessian = Bk
    return state1


def diag_sr1_update_(
        state1: OptimisationState, state: OptimisationState, tol=1e-8, **kwargs
) -> OptimisationState:
    yk = VariableData.sub(state1.gradient, state.gradient)
    dk = VariableData.sub(state1.parameters, state.parameters)
    Bk = state.hessian
    zk = yk + Bk * dk
    dzk = dk * zk
    # alpha = -zk.dot(dk) / dzk.dot(dzk)
    alpha = -(zk * dk).var_sum()
    tols = tol * dk.var_norm() ** 2 * zk.var_norm() ** 2
    for v, d in (dzk ** 2).var_sum().items():
        if d > tols[v]:
            alpha[v] /= d
        else:
            alpha[v] = 0.0

    Bk = Bk.diagonalupdate(alpha * (zk ** 2))

    state1.hessian = Bk
    return state1


def diag_sr1_bfgs_update(
        state1: OptimisationState, state: OptimisationState, **kwargs
) -> OptimisationState:
    yk = VariableData.sub(state1.gradient, state.gradient)
    dk = VariableData.sub(state1.parameters, state.parameters)
    Bk = state.hessian
    zk = yk + Bk * dk
    dzk = dk * zk


def bfgs1_update(
        state1: OptimisationState,
        state: OptimisationState,
        **kwargs,
) -> OptimisationState:
    """
    y_k = g_{k+1} - g{k}
    d_k = x_{k+1} - x{k}
    B_{k+1} = B_{k}
    + \frac
        {y_{k}y_{k}^T}
        {y_{k}^T d_{k}}}
    - \frac
        {B_{k} d_{k} (B_{k} d_{k})^T}
        {d_{k}^T B_{k} d_{k}}}}
    """
    yk = VariableData.sub(state.gradient, state1.gradient)
    dk = VariableData.sub(state1.parameters, state.parameters)
    Bk = state.hessian

    ykTdk = yk.dot(dk)
    Bdk = Bk.dot(dk)
    dkTBdk = -VariableData.dot(Bdk, dk)

    state1.hessian = Bk.update(
        (yk, VariableData(yk).div(ykTdk)), (Bdk, VariableData(Bdk).div(dkTBdk))
    )
    return state1


def bfgs_update(
        state1: OptimisationState,
        state: OptimisationState,
        **kwargs,
) -> OptimisationState:
    yk = VariableData.sub(state1.gradient, state.gradient)
    dk = VariableData.sub(state1.parameters, state.parameters)
    Bk = state.hessian

    ykTdk = -yk.dot(dk)
    Bdk = Bk.dot(dk)
    dkTBdk = VariableData.dot(Bdk, dk)

    state1.hessian = Bk.update(
        (yk, VariableData(yk).div(ykTdk)), (Bdk, VariableData(Bdk).div(dkTBdk))
    )
    return state1


def quasi_deterministic_update(
        state1: OptimisationState,
        state: OptimisationState,
        **kwargs,
) -> OptimisationState:
    dk = VariableData.sub(state1.parameters, state.parameters)
    zk = VariableData.sub(
        state1.value.deterministic_values, state.value.deterministic_values
    )
    Bxk, Bzk = state1.hessian, state.det_hessian
    zkTzk2 = zk.dot(zk) ** 2
    alpha = (dk.dot(Bxk.dot(dk)) - zk.dot(Bzk.dot(zk))) / zkTzk2
    if alpha >= 0:
        Bzk1 = Bzk.lowrankupdate(np.sqrt(alpha) * (zk))
    else:
        Bzk1 = Bzk.lowrankdowndate(np.sqrt(-alpha) * (zk))

    state1.det_hessian = Bzk1
    return state1


def diag_quasi_deterministic_update(
        state1: OptimisationState,
        state: OptimisationState,
        **kwargs,
) -> OptimisationState:
    dk = VariableData.sub(state1.parameters, state.parameters)
    zk = VariableData.sub(
        state1.value.deterministic_values, state.value.deterministic_values
    )
    Bxk, Bzk = state1.hessian, state.det_hessian
    zk2 = zk ** 2
    zk4 = (zk2 ** 2).sum()
    alpha = (dk.dot(Bxk.dot(dk)) - zk.dot(Bzk.dot(zk))) / zk4
    state1.det_hessian = Bzk.diagonalupdate(alpha * zk2)

    return state1


class QuasiNewtonUpdate:
    def __init__(self, quasi_newton_update, det_quasi_newton_update):
        self.quasi_newton_update = quasi_newton_update
        self.det_quasi_newton_update = det_quasi_newton_update

    def __call__(
            self,
            state1: OptimisationState,
            state: OptimisationState,
            **kwargs,
    ) -> OptimisationState:

        # Only update estimate if a step has been taken
        state1 = self.quasi_newton_update(state1, state, **kwargs)
        if state.det_hessian:
            state1 = self.det_quasi_newton_update(state1, state, **kwargs)

        return state1


full_bfgs_update = QuasiNewtonUpdate(bfgs_update, quasi_deterministic_update)
full_sr1_update = QuasiNewtonUpdate(sr1_update, quasi_deterministic_update)
full_diag_update = QuasiNewtonUpdate(diag_sr1_update, diag_quasi_deterministic_update)


## Newton step


def take_step(
        state: OptimisationState,
        old_state: Optional[OptimisationState] = None,
        *,
        search_direction=newton_abs_direction,
        calc_line_search=line_search,
        search_direction_kws: Optional[Dict[str, Any]] = None,
        line_search_kws: Optional[Dict[str, Any]] = None,
) -> Tuple[Optional[float], OptimisationState]:
    state.search_direction = search_direction(state, **(search_direction_kws or {}))
    return calc_line_search(state, old_state, **(line_search_kws or {}))


def take_quasi_newton_step(
        state: OptimisationState,
        old_state: Optional[OptimisationState] = None,
        *,
        search_direction=newton_abs_direction,
        calc_line_search=line_search,
        quasi_newton_update=full_bfgs_update,
        search_direction_kws: Optional[Dict[str, Any]] = None,
        line_search_kws: Optional[Dict[str, Any]] = None,
        quasi_newton_kws: Optional[Dict[str, Any]] = None,
) -> Tuple[Optional[float], OptimisationState]:
    """ """
    state.search_direction = search_direction(state, **(search_direction_kws or {}))
    if state.search_direction.vecnorm(np.Inf) == 0:
        # if gradient is zero then at maximum already
        return 0.0, state

    stepsize, state1 = calc_line_search(state, old_state, **(line_search_kws or {}))
    if stepsize:
        # Only update estimate if a step has been taken
        state1 = quasi_newton_update(state1, state, **(quasi_newton_kws or {}))

    return stepsize, state1


def xtol_condition(state, old_state, xtol=1e-6, ord=None, **kwargs):
    if not old_state:
        return

    dx = VariableData.sub(state.parameters, old_state.parameters).vecnorm(ord=ord)
    if dx < xtol:
        return True, f"Minimum parameter change tolerance achieved, {dx} < {xtol}"


def grad_condition(state, old_state, gtol=1e-5, ord=None, **kwargs):
    dg = VariableData.vecnorm(state.gradient, ord=ord)
    if dg < gtol:
        return True, f"Gradient tolerance achieved, {dg} < {gtol}"


def ftol_condition(state, old_state, ftol=1e-6, monotone=True, **kwargs):
    if not old_state:
        return

    df = state.value - old_state.value
    if 0 < df < ftol:
        return True, f"Minimum function change tolerance achieved, {df} < {ftol}"
    elif monotone and df < 0:
        return False, f"factor failed to increase on next step, {df}"


def nfev_condition(state, old_state, maxfev=10000, **kwargs):
    if state.f_count > maxfev:
        return (
            True,
            f"Maximum number of function evaluations (maxfev={maxfev}) has been exceeded.",
        )


def ngev_condition(state, old_state, maxgev=10000, **kwargs):
    if state.g_count > maxgev:
        return (
            True,
            f"Maximum number of gradient evaluations (maxgev={maxgev}) has been exceeded.",
        )


stop_conditions = (
    xtol_condition,
    ftol_condition,
    grad_condition,
)

_OPT_CALLBACK = Callable[[OptimisationState, OptimisationState], None]


def check_stop_conditions(
        stepsize, state, old_state, stop_conditions, **stop_kws
) -> Optional[Tuple[bool, str]]:
    if stepsize is None:
        return False, "abnormal termination of line search"
    elif not np.isfinite(state.value):
        return False, "function is no longer finite"
    else:
        for stop_condition in stop_conditions:
            stop = stop_condition(state, old_state, **(stop_kws or {}))
            if stop:
                return stop


def optimise_quasi_newton(
        state: OptimisationState,
        old_state: Optional[OptimisationState] = None,
        *,
        max_iter=100,
        search_direction=newton_abs_direction,
        calc_line_search=line_search,
        quasi_newton_update=bfgs_update,
        stop_conditions=stop_conditions,
        search_direction_kws: Optional[Dict[str, Any]] = None,
        line_search_kws: Optional[Dict[str, Any]] = None,
        quasi_newton_kws: Optional[Dict[str, Any]] = None,
        stop_kws: Optional[Dict[str, Any]] = None,
        callback: Optional[_OPT_CALLBACK] = None,
        **kwargs,
) -> Tuple[OptimisationState, Status]:
    success = True
    updated = False
    messages = ()
    message = "max iterations reached"
    stepsize = 0.0
    for i in range(max_iter):
        stop = check_stop_conditions(
            stepsize, state, old_state, stop_conditions, **(stop_kws or {})
        )
        if stop:
            success, message = stop
            break

        with LogWarnings(logger=_log_projection_warnings, action='always') as caught_warnings:
            try:
                stepsize, state1 = take_quasi_newton_step(
                    state,
                    old_state,
                    search_direction=search_direction,
                    calc_line_search=calc_line_search,
                    quasi_newton_update=quasi_newton_update,
                    search_direction_kws=search_direction_kws,
                    line_search_kws=line_search_kws,
                    quasi_newton_kws=quasi_newton_kws,
                )
            except exc.MessageException as e:
                logger.exception(e)
                success = False
                message = f"Line search failed due to {e}"
                break

        for m in caught_warnings.messages:
            messages += (f"optimise_quasi_newton warning: {m}",)

        if stepsize is None:
            success = False
            message = "Line search failed"
            break

        updated = True
        state, old_state = state1, state
        i += 1
        if callback:
            callback(state, old_state)

    message += f", iter={i}"
    messages += (message,)
    status = Status(
        success,
        messages=messages,
        updated=updated,
        flag=StatusFlag.get_flag(success, i),
    )
    return state, status



import csv
import json
import warnings
from copy import copy
from typing import List, Optional, Tuple, Union

import numpy as np

from autofit import exc
from autofit.mapper.model import ModelInstance
from autofit.mapper.prior_model.abstract import AbstractPriorModel, Path
from autofit.non_linear.samples.sample import Sample


class Samples:
    def __init__(
            self,
            model: AbstractPriorModel,
            sample_list: List[Sample],
            total_iterations: Optional[int] = None,
            time: Optional[float] = None,
            results_internal: Optional = None,
    ):
        """
        The `Samples` classes in **PyAutoFit** provide an interface between the results_internal of
        a `NonLinearSearch` (e.g. as files on your hard-disk) and Python.

        For example, the output class can be used to load an instance of the best-fit model, get an instance of any
        individual sample by the `NonLinearSearch` and return information on the likelihoods, errors, etc.

        This class stores samples of searches which provide maximum likelihood estimates of the  model-fit (e.g.
        PySwarms, LBFGS).

        To use a library's in-built visualization tools results are optionally stored in their native internal format
        using the `results_internal` attribute.

        Parameters
        ----------
        model
            Maps input vectors of unit parameter values to physical values and model instances via priors.
        sample_list
            The list of `Samples` which contains the paramoeters, likelihood, weights, etc. of every sample taken
            by the non-linear search.
        total_iterations
            The total number of iterations, which often cannot be estimated from the sample list (which contains
            only accepted samples).
        time
            The time taken to perform the model-fit, which is passed around `Samples` objects for outputting
            information on the overall fit.
        results_internal
            The nested sampler's results in their native internal format for interfacing its visualization library.
        """
        self.model = model
        self.sample_list = sample_list

        self.total_iterations = total_iterations
        self.time = time
        self.results_internal = results_internal

        self._paths = None
        self._names = None

    def __add__(
            self,
            other: "Samples"
    ) -> "Samples":
        """
        Samples can be added together, which combines their `sample_list` meaning that inferred parameters are
        computed via their joint PDF.

        Parameters
        ----------
        other
            The Samples to be added to this Samples instance.

        Returns
        -------
        A class that combined the samples of the two Samples objects.
        """

        self._check_addition(other=other)

        if self.results_internal is not None:
            warnings.warn(
                f"Addition of {self.__class__.__name__} cannot retain results in native format. "
                "Visualization of summed samples diabled.",
                exc.SamplesWarning
            )

        return self.__class__(
            model=self.model,
            sample_list=self.sample_list + other.sample_list,
            time=self.time
        )

    def __radd__(self, other):
        """
        Samples can be added together, which combines their `sample_list` meaning that inferred parameters are
        computed via their joint PDF.

        Overwriting `__radd__` enables the sum function to be used on a list of samples, e.g.:

        `samples = sum([samples_x5, samples_x5, samples_x5])`

        Parameters
        ----------
        other
            The Samples to be added to this Samples instance.

        Returns
        -------
        A class that combines the samples of a list of Samples objects.
        """
        return self

    def _check_addition(self, other: "Samples"):
        """
        When adding samples together, perform the following checks to make sure it is valid to add the two objects
        together:

        - That both objects being added are `Samples` objects.
        - That both models have the same prior count, else the dimensionality does not allow for valid addition.
        - That both `Samples` objects use an identical model, such that we are adding together the same parameters.

        Parameters
        ----------
        other
            The Samples to be added to this Samples instance.
        """

        def raise_exc():
            raise exc.SamplesException(
                "Cannot add together two Samples objects which have different models."
            )

        if not isinstance(
                self,
                Samples
        ):

            raise_exc()

        if not isinstance(
                other,
                Samples
        ):
            raise_exc()

        if self.model.prior_count != other.model.prior_count:
            raise_exc()

        for path_self, path_other in zip(self.model.paths, other.model.paths):
            if path_self != path_other:
                raise_exc()

    def values_for_path(
            self,
            path: Tuple[str]
    ) -> List[float]:
        """
        Returns the value for a variable with a given path
        for each sample in the model
        """
        return [
            sample.kwargs[
                path
            ]
            for sample
            in self.sample_list
        ]

    @property
    def log_evidence(self) -> Optional[float]:
        return None

    @property
    def paths(self) -> List[Tuple[Path]]:
        """
        A list of paths to unique priors in the same order as prior
        ids (and therefore sample columns)

        Uses hasattr to make backwards compatible
        """
        if not hasattr(self, "_paths") or self._paths is None:
            self._paths = self.model.all_paths
        return self._paths

    @property
    def names(self) -> List[Tuple[str]]:
        """
        A list of names of unique priors in the same order as prior
        ids (and therefore sample columns)

        Uses hasattr to make backwards compatible
        """
        if not hasattr(self, "_names") or self._names is None:
            self._names = self.model.all_names
        return self._names

    @property
    def parameter_lists(self):
        result = list()
        for sample in self.sample_list:
            tuples = self.paths if sample.is_path_kwargs else self.names
            result.append(
                sample.parameter_lists_for_paths(
                    tuples
                )
            )

        return result

    @property
    def total_samples(self):
        return len(self.sample_list)

    def __len__(self):
        return len(self.sample_list)

    @property
    def weight_list(self):
        return [
            sample.weight
            for sample
            in self.sample_list
        ]

    @property
    def log_likelihood_list(self):
        return [
            sample.log_likelihood
            for sample
            in self.sample_list
        ]

    @property
    def log_posterior_list(self):
        return [
            sample.log_posterior
            for sample
            in self.sample_list
        ]

    @property
    def log_prior_list(self):
        return [
            sample.log_prior
            for sample
            in self.sample_list
        ]

    @property
    def parameters_extract(self):
        return np.asarray(self.parameter_lists).T

    @property
    def _headers(self) -> List[str]:
        """
        Headers for the samples table
        """

        return self.model.model_component_and_parameter_names + [
            "log_likelihood",
            "log_prior",
            "log_posterior",
            "weight",
        ]

    @property
    def _rows(self) -> List[List[float]]:
        """
        Rows in the samples table
        """

        log_likelihood_list = self.log_likelihood_list
        log_prior_list = self.log_prior_list
        log_posterior_list = self.log_posterior_list
        weight_list = self.weight_list

        for index, row in enumerate(self.parameter_lists):
            yield row + [
                log_likelihood_list[index],
                log_prior_list[index],
                log_posterior_list[index],
                weight_list[index],
            ]

    def write_table(self, filename: str):
        """
        Write a table of parameters, posteriors, priors and likelihoods.

        Parameters
        ----------
        filename
            Where the table is to be written
        """

        with open(filename, "w+", newline="") as f:
            writer = csv.writer(f)
            writer.writerow(self._headers)
            for row in self._rows:
                writer.writerow(row)

    @property
    def info_json(self):
        return {}

    def info_to_json(self, filename):
        with open(filename, 'w') as outfile:
            json.dump(self.info_json, outfile)

    @property
    def max_log_likelihood_sample(self) -> Sample:
        """
        The index of the sample with the highest log likelihood.
        """
        most_likely_sample = None
        for sample in self.sample_list:
            if most_likely_sample is None or sample.log_likelihood > most_likely_sample.log_likelihood:
                most_likely_sample = sample
        return most_likely_sample

    @property
    def max_log_posterior_sample(self) -> Sample:
        return self.sample_list[
            self.max_log_posterior_index
        ]

    @property
    def max_log_likelihood_vector(self) -> List[float]:
        """
        The parameters of the maximum log likelihood sample of the `NonLinearSearch` returned as a list of values.
        """
        sample = self.max_log_likelihood_sample
        return sample.parameter_lists_for_paths(
            self.paths if sample.is_path_kwargs else self.names
        )

    @property
    def max_log_likelihood_instance(self) -> ModelInstance:
        """
        The parameters of the maximum log likelihood sample of the `NonLinearSearch` returned as a model instance.
        """
        return self.max_log_likelihood_sample.instance_for_model(
            self.model
        )

    @property
    def max_log_posterior_index(self) -> int:
        """
        The index of the sample with the highest log posterior.
        """
        return int(np.argmax(self.log_posterior_list))

    @property
    def max_log_posterior_vector(self) -> List[float]:
        """
        The parameters of the maximum log posterior sample of the `NonLinearSearch` returned as a list of values.
        """
        return self.parameter_lists[self.max_log_posterior_index]

    @property
    def max_log_posterior_instance(self) -> ModelInstance:
        """
        The parameters of the maximum log posterior sample of the `NonLinearSearch` returned as a model instance.
        """
        return self.model.instance_from_vector(
            vector=self.max_log_posterior_vector,
        )

    def gaussian_priors_at_sigma(self, sigma: float) -> [List]:
        """
        `GaussianPrior`s of every parameter used to link its inferred values and errors to priors used to sample the
        same (or similar) parameters in a subsequent search, where:

        - The mean is given by maximum log likelihood model values.
        - Their errors are omitted, as this information is not available from an search. When these priors are
        used to link to another search, it will thus automatically use the prior config values.

        Parameters
        -----------
        sigma
            The sigma limit within which the PDF is used to estimate errors (e.g. sigma = 1.0 uses 0.6826 of the PDF).
        """
        return list(map(lambda vector: (vector, 0.0), self.max_log_likelihood_vector))

    def instance_from_sample_index(self, sample_index) -> ModelInstance:
        """
        The parameters of an individual sample of the non-linear search, returned as a model instance.

        Parameters
        -----------
        sample_index
            The sample index of the weighted sample to return.
        """
        return self.model.instance_from_vector(vector=self.parameter_lists[sample_index])

    def minimise(self) -> "Samples":
        """
        A copy of this object with only important samples retained
        """
        samples = copy(self)
        samples.model = None
        samples.sample_list = list({
            self.max_log_likelihood_sample,
            self.max_log_posterior_sample
        })
        return samples

    def with_paths(
            self,
            paths: Union[List[Tuple[str, ...]], List[str]]
    ) -> "Samples":
        """
        Create a copy of this object with only attributes specified
        by a list of paths.

        Parameters
        ----------
        paths
            A list of paths to attributes. Only kwargs and model components
            specified by these paths are retained.

            All children of a given path are retained.

        Returns
        -------
        A set of samples with a reduced set of attributes
        """
        with_paths = copy(self)
        with_paths.model = self.model.with_paths(
            paths
        )
        with_paths.sample_list = [
            sample.with_paths(paths)
            for sample in self.sample_list
        ]
        return with_paths

    def __setstate__(self, state):

        self.__dict__.update(state)

        # This is a hack to fix BC on some results implemented 25/01/22, safe to delete in a month or so...

        try:
            if "results_internal" not in state:
                self.results_internal = state["results"]
        except KeyError:
            pass

    def __copy__(self):
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)
        result._names = None
        result._paths = None
        return result

    def without_paths(
            self,
            paths: Union[List[Tuple[str, ...]], List[str]]
    ) -> "Samples":
        """
        Create a copy of this object with only attributes not specified
        by a list of paths.

        Parameters
        ----------
        paths
            A list of paths to attributes. kwargs and model components
            specified by these paths are removed.

            All children of a given path are removed.

        Returns
        -------
        A set of samples with a reduced set of attributes
        """
        with_paths = copy(self)
        with_paths.model = self.model.without_paths(
            paths
        )
        with_paths.sample_list = [
            sample.without_paths(paths)
            for sample in self.sample_list
        ]
        return with_paths

    def subsamples(self, model):
        if self.model is None:
            return None

        path_map = {
            tuple(self.model.all_paths_for_prior(
                prior
            )): path
            for path, prior in model.path_priors_tuples
        }
        copied = copy(self)
        copied._paths = None
        copied._names = None
        copied.model = model

        copied.sample_list = [
            sample.subsample(path_map)
            for sample in self.sample_list
        ]
        return copied

    @property
    def median_pdf_vector(self) -> ModelInstance:
        raise NotImplementedError

    @property
    def median_pdf_instance(self) -> ModelInstance:
        raise NotImplementedError

    def vector_at_sigma(self, sigma: float) -> [(float, float)]:
        raise NotImplementedError

    def vector_at_upper_sigma(self, sigma: float) -> [float]:
        raise NotImplementedError

    def vector_at_lower_sigma(self, sigma: float) -> [float]:
        raise NotImplementedError

    def instance_at_sigma(self, sigma: float) -> ModelInstance:
        raise NotImplementedError

    def instance_at_upper_sigma(self, sigma: float) -> ModelInstance:
        raise NotImplementedError

    def instance_at_lower_sigma(self, sigma: float) -> ModelInstance:
        raise NotImplementedError

    def error_vector_at_sigma(self, sigma: float) -> [(float, float)]:
        raise NotImplementedError

    def error_vector_at_upper_sigma(self, sigma: float) -> [float]:
        raise NotImplementedError

    def error_vector_at_lower_sigma(self, sigma: float) -> [float]:
        raise NotImplementedError

    def error_magnitude_vector_at_sigma(self, sigma: float) -> [float]:
        raise NotImplementedError

    def error_instance_at_sigma(self, sigma: float) -> ModelInstance:
        raise NotImplementedError

    def error_instance_at_upper_sigma(self, sigma: float) -> ModelInstance:
        raise NotImplementedError

    def error_instance_at_lower_sigma(self, sigma: float) -> ModelInstance:
        raise NotImplementedError


from typing import List

import numpy as np

from autofit import exc
from autofit.mapper import model_mapper as mm
from autofit.mapper.prior.abstract import Prior
from autofit.non_linear.result import Result, Placeholder

LimitLists = List[List[float]]


class GridSearchResult:

    def __init__(
            self,
            results: List[Result],
            lower_limits_lists: LimitLists,
            grid_priors: List[Prior]
    ):
        """
        The result of a grid search.

        Parameters
        ----------
        results
            The results of the non linear optimizations performed at each grid step
        lower_limits_lists
            A list of lists of values representing the lower bounds of the grid searched values at each step
        """
        self.lower_limits_lists = lower_limits_lists
        self.results = results
        self.no_dimensions = len(self.lower_limits_lists[0])
        self.no_steps = len(self.lower_limits_lists)
        self.side_length = int(self.no_steps ** (1 / self.no_dimensions))
        self.step_size = 1 / self.side_length
        self.grid_priors = grid_priors

    @property
    def physical_lower_limits_lists(self) -> LimitLists:
        """
        The lower physical values for each grid square
        """
        return self._physical_values_for(
            self.lower_limits_lists
        )

    @property
    def physical_centres_lists(self) -> LimitLists:
        """
        The middle physical values for each grid square
        """
        return self._physical_values_for(
            self.centres_lists
        )

    @property
    def physical_upper_limits_lists(self) -> LimitLists:
        """
        The upper physical values for each grid square
        """
        return self._physical_values_for(
            self.upper_limits_lists
        )

    @property
    def upper_limits_lists(self) -> LimitLists:
        """
        The upper values for each grid square
        """
        return [
            [
                limit + self.step_size
                for limit in limits
            ]
            for limits in self.lower_limits_lists
        ]

    @property
    def centres_lists(self) -> LimitLists:
        """
        The centre values for each grid square
        """
        return [
            [
                (upper + lower) / 2
                for upper, lower
                in zip(upper_limits, lower_limits)
            ]
            for upper_limits, lower_limits in zip(
                self.lower_limits_lists,
                self.upper_limits_lists
            )
        ]

    def _physical_values_for(
            self,
            unit_lists: LimitLists
    ) -> LimitLists:
        """
        Compute physical values for lists of lists of unit hypercube
        values.

        Parameters
        ----------
        unit_lists
            A list of lists of hypercube values

        Returns
        -------
        A list of lists of physical values
        """
        return [
            [
                prior.value_for(
                    limit
                )
                for prior, limit in
                zip(
                    self.grid_priors,
                    limits
                )
            ]
            for limits in unit_lists
        ]

    def __setstate__(self, state):
        return self.__dict__.update(state)

    def __getstate__(self):
        return self.__dict__

    def __getattr__(self, item: str) -> object:
        """
        We default to getting attributes from the best result. This allows promises to reference best results.
        """
        return getattr(self.best_result, item)

    @property
    def shape(self):
        return self.no_dimensions * (int(self.no_steps ** (1 / self.no_dimensions)),)

    @property
    def best_result(self):
        """
        The best result of the grid search. That is, the result output by the non linear search that had the highest
        maximum figure of merit.

        Returns
        -------
        best_result: Result
        """
        best_result = Placeholder()
        for result in self.results:
            if result > best_result:
                best_result = result
        return best_result

    @property
    def best_model(self):
        """
        Returns
        -------
        best_model: mm.ModelMapper
            The model mapper instance associated with the highest figure of merit from the grid search
        """
        return self.best_result.model

    @property
    def all_models(self):
        """
        Returns
        -------
        all_models: [mm.ModelMapper]
            All model mapper instances used in the grid search
        """
        return [result.model for result in self.results]

    @property
    def physical_step_sizes(self):

        physical_step_sizes = []

        # TODO : Make this work for all dimensions in a less ugly way.

        for dim in range(self.no_dimensions):

            values = [value[dim] for value in self.physical_lower_limits_lists]
            diff = [abs(values[n] - values[n - 1]) for n in range(1, len(values))]

            if dim == 0:
                physical_step_sizes.append(np.max(diff))
            elif dim == 1:
                physical_step_sizes.append(np.min(diff))
            else:
                raise exc.GridSearchException(
                    "This feature does not support > 2 dimensions"
                )

        return tuple(physical_step_sizes)

    def _list_to_native(self, lst: List):
        return np.reshape(np.array(lst), self.shape)

    @property
    def results_native(self):
        """
        The result of every grid search on a NumPy array whose shape is the native dimensions of the grid search.

        For example, for a 2x2 grid search the shape of the Numpy array is (2,2) and it is numerically ordered such
        that the first search's result (corresponding to unit priors (0.0, 0.0)) are in the first value (E.g. entry
        [0, 0]) of the NumPy array.
        """
        return self._list_to_native(lst=[result for result in self.results])

    @property
    def log_likelihoods_native(self):
        """
        The maximum log likelihood of every grid search on a NumPy array whose shape is the native dimensions of the
        grid search.

        For example, for a 2x2 grid search the shape of the Numpy array is (2,2) and it is numerically ordered such
        that the first search's maximum likelihood (corresponding to unit priors (0.0, 0.0)) are in the first
        value (E.g. entry [0, 0]) of the NumPy array.
        """
        return self._list_to_native(lst=[result.log_likelihood for result in self.results])

    @property
    def log_evidences_native(self):
        """
        The log evidence of every grid search on a NumPy array whose shape is the native dimensions of the grid search.

        For example, for a 2x2 grid search the shape of the Numpy array is (2,2) and it is numerically ordered such
        that the first search's log evidence (corresponding to unit priors (0.0, 0.0)) are in the first value (E.g.
        entry [0, 0]) of the NumPy array.
        """
        return self._list_to_native(lst=[result.samples.log_evidence for result in self.results])



from inspect import getfullargspec
from typing import Tuple, Dict, Any, Callable, Union, List, Optional, TYPE_CHECKING

import numpy as np
try:
    import jax

    _HAS_JAX = True
except ImportError:
    _HAS_JAX = False

from autofit.graphical.utils import (
    nested_filter,
    is_variable,
    try_getitem, 
)
from autofit.mapper.variable import Variable, Plate, VariableData


from autofit.graphical.factor_graphs.abstract import FactorValue, AbstractFactor


if TYPE_CHECKING:
    from autofit.graphical.factor_graphs.jacobians import (
        VectorJacobianProduct, JacobianVectorProduct
    )


class Factor(AbstractFactor):
    """Represents factors in Graphical models. The functions passed to this
    object will be called by positional arguments (to allow compatibility)
    with the jax API.

    Parameters
    ----------
    factor
        the function being wrapped, must accept calls
        through positional arguments

    *args: Variables
        Variables for each positional argument for the function

    factor_out, default FactorValue:
        The output of the factor. This can just be `FactorValue`
        or can be a arbitrarily nested structure of lists, tuples and dicts
        e.g.
        >>> foo = lambda x, y: (z, {'a': [a]})
        >>> factor = Factor(foo, x, y, factor_out=(z, {'a': [a]}))
    name: optional, str
        the name of the factor, if not passed then uses the name
        of the function passed

    plates: Tuple[Plate, ...] = ()
        plates that the factor are associated with

    vjp: optional False
        if True uses jax.vjp to calculate the Jacobian of the
        outputs

    factor_vjp: optional
        Must be a function produces functionaly equivalent
        output to jax.vjp(factor, *args) <equiv> factor_vjp(*args)

    factor_jacobian: optional
        function equivalent to calling,
        factor(*args), jax.jacobian(factor, len(range(args)))(*args)

    jacobian=None,
        function equivalent to calling,
        jax.jacobian(factor, len(range(args)))

    numerical_jacobian=True
        if True calculates Jacobian using finite differences
        if False calculates Jacobian using jax

    jacfwd=True
        if calculates jacobian using jax.jacfwd instead of
        jax.jacobian

    eps=1e-8
        the interval overwhich to calculate the finite differences




    Methods
    -------
    __call__({x: x0}) -> FactorValue
        calls the factor, the passed input must be a dictionary with
        where the keys are the Variable objects that the function takes
        as input.

        returns a FactorValue object which behaves like an np.ndarray
        deterministic values are stored in the deterministic_values
        attribute

    func_jacobian({x: x0}) -> Tuple[FactorValue, AbstractJacobianValue]
        calls the factor and returns it value and the Jacobian of its value
        with respect to the `variables` passed. The Jacobian is returned as
        a VariableLinearOperator with the appropriate methods for calculating
        the vector-Jacobian or Jacobian-vector products depending on how
        the Jacobian is calculated internally.

    Examples
    --------
    def linear(x, a, b):
        z = x.dot(a) + b
        return (z**2).sum(), z

    def likelihood(y, z):
        return ((y - z)**2).sum()

    def combined(x, y, a, b):
        like, z = linear(x, a, b)
        return like + likelihood(y, z)

    x_, a_, b_, y_, z_ = variables("x, a, b, y, z")
    x = np.arange(10.).reshape(5, 2)
    a = np.arange(2.).reshape(2, 1)
    b = np.ones(1)
    y = np.arange(0., 10., 2).reshape(5, 1)
    values = {x_: x, y_: y, a_: a, b_: b}
    linear_factor = FactorVJP(
        linear, x_, a_, b_, factor_out=(FactorValue, z_))
    like_factor = FactorVJP(likelihood, y_, z_)
    full_factor = FactorVJP(combined, x_, y_, a_, b_)

    x = np.arange(10.).reshape(5, 2)
    a = np.arange(2.).reshape(2, 1)
    b = np.ones(1)
    y = np.arange(0., 10., 2).reshape(5, 1)
    values = {x_: x, y_: y, a_: a, b_: b}

    # Fully working problem
    fval, jac = full_factor.func_jacobian(values)
    grad = jac(1)

    linear_val, linear_jac = linear_factor.func_jacobian(values)
    like_val, like_jac = like_factor.func_jacobian(
        {**values, **linear_val.deterministic_values})
    combined_val = like_val + linear_val

    combined_grads = {FactorValue: 1.}
    for v, g in like_jac(combined_grads).items():
        combined_grads[v] = g + combined_grads.get(v, 0)

    for v, g in linear_jac(combined_grads).items():
        combined_grads[v] = g + combined_grads.get(v, 0)

    assert (fval.log_value - combined_val.log_value) == 0
    assert (grad - combined_grads).norm() == 0
    """

    def __init__(
            self,
            factor,
            *args: Variable,
            name="",
            arg_names=None,
            factor_out=FactorValue,
            plates: Tuple[Plate, ...] = (),
            vjp=False,
            factor_vjp=None,
            factor_jacobian=None,
            jacobian=None,
            numerical_jacobian=True,
            jacfwd=True,
            eps=1e-8,
    ):
        if not arg_names:
            arg_names = [arg for arg in getfullargspec(factor).args]
            if arg_names and arg_names[0] == "self":
                arg_names = arg_names[1:]

        # Make sure arg_names matches length of args
        for v in args[len(arg_names):]:
            arg_name = v.name
            # Make sure arg_name is unique
            while arg_name in arg_names:
                arg_name += "_"
            arg_names.append(arg_name)

        kwargs = dict(zip(arg_names, args))
        name = name or factor.__name__

        AbstractFactor.__init__(
            self,
            name=name,
            plates=plates,
            factor_out=factor_out,
            **kwargs,
        )

        # self.factor_out = factor_out
        self.eps = eps
        self._set_factor(factor)
        self._set_jacobians(
            vjp=vjp,
            factor_vjp=factor_vjp,
            factor_jacobian=factor_jacobian,
            jacobian=jacobian,
            numerical_jacobian=numerical_jacobian,
            jacfwd=jacfwd,
        )

    @property
    def shape(self) -> Tuple[int, ...]:
        if self.plates:
            return self._factor.shape

        return ()

    def __getitem__(
        self, plates_index: Dict[Plate, Union[List[int], range, slice]]
    ) -> "Factor":
        return self.subset(plates_index)

    def subset(
        self,
        plates_index: Dict[Plate, Union[List[int], range, slice]],
        plate_sizes: Optional[Dict[Plate, int]] = None,
    ) -> "Factor":
        if not self.plates:
            return self

        plate_sizes = plate_sizes or dict(zip(self.plates, self.shape))
        index = Variable.make_indexes(self, plates_index, plate_sizes)
        subset_factor = self._factor[index]
        kws = self._subset_jacobian(subset_factor, index)

        subset = ", ".join(map("{0.name}={1}".format, self.plates, map(np.size, index)))
        kws["name"] = f"{self.name}[{subset}]"
        kws["eps"] = self.eps
        kws["factor_out"] = self.factor_out
        kws["plates"] = self.plates
        kws["arg_names"] = self.arg_names

        return Factor(subset_factor, *self.args, **kws)

    def _subset_jacobian(self, subset_factor, index):
        jac_kws = {"vjp": self._vjp}
        if self._vjp:
            factor_vjp = None
            if self._factor_vjp != self._jax_factor_vjp:
                factor_vjp = try_getitem(
                    self._factor_vjp, index, getattr(subset_factor, 'factor_vjp', None)
                )

            jac_kws["factor_vjp"] = factor_vjp
        else:
            factor_jacobian = None
            jacobian = None
            if self._factor_jacobian != Factor._factor_jacobian:
                factor_jacobian = try_getitem(
                    self._factor_jacobian, index, getattr(subset_factor, 'factor_jacobian', None)
                )
            elif self._jacobian != Factor._jacobian:
                jacobian = try_getitem(
                    self.jacobian, index, getattr(subset_factor, 'jacobian', None)
                )
            elif self._factor_jacobian != self._numerical_factor_jacobian:
                jac_kws["jacfwd"] = self._jacfwd

            jac_kws["jacobian"] = jacobian
            jac_kws["factor_jacobian"] = factor_jacobian

        return jac_kws

    def _set_jacobians(
            self,
            vjp=False,
            factor_vjp=None,
            factor_jacobian=None,
            jacobian=None,
            numerical_jacobian=True,
            jacfwd=True,
    ):
        self._vjp = vjp
        self._jacfwd = jacfwd
        if vjp or factor_vjp:
            if factor_vjp:
                self._factor_vjp = factor_vjp
            elif not _HAS_JAX:
                raise ModuleNotFoundError(
                    "jax needed if `factor_vjp` not passed with vjp=True"
                )

            self.func_jacobian = self._vjp_func_jacobian
        else:
            # This is set by default
            # self.func_jacobian = self._jvp_func_jacobian

            if factor_jacobian:
                self._factor_jacobian = factor_jacobian
            elif jacobian:
                self._jacobian = jacobian
            elif numerical_jacobian:
                self._factor_jacobian = self._numerical_factor_jacobian
            elif _HAS_JAX:
                if jacfwd:
                    self._jacobian = jax.jacfwd(self._factor, range(self.n_args))
                else:
                    self._jacobian = jax.jacobian(self._factor, range(self.n_args))

    def _factor_value(self, raw_fval) -> FactorValue:
        """Converts the raw output of the factor into a `FactorValue`
        where the values of the deterministic values are stored in a dict
        attribute `FactorValue.deterministic_values`
        """
        det_values = VariableData(nested_filter(is_variable, self.factor_out, raw_fval))
        fval = det_values.pop(FactorValue, 0.0)
        return FactorValue(fval, det_values)

    def __call__(self, values: VariableData) -> FactorValue:
        """Calls the factor with the values specified by the dictionary of
        values passed, returns a FactorValue with the value returned by the
        factor, and any deterministic factors"""
        raw_fval = self._factor_args(*(values[v] for v in self.args))
        return self._factor_value(raw_fval)

    def _jax_factor_vjp(self, *args) -> Tuple[Any, Callable]:
        return jax.vjp(self._factor, *args)

    _factor_vjp = _jax_factor_vjp

    def _vjp_func_jacobian(
            self, values: VariableData
    ) -> Tuple[FactorValue, "VectorJacobianProduct"]:
        """Calls the factor and returns the factor value with deterministic
        values, and a `VectorJacobianProduct` operator that allows the
        calculation of the gradient of the input values to be calculated
        with respect to the gradients of the output values (i.e backprop)
        """
        from autofit.graphical.factor_graphs.jacobians import (
            VectorJacobianProduct,
        )
        raw_fval, fvjp = self._factor_vjp(*(values[v] for v in self.args))
        fval = self._factor_value(raw_fval)

        fvjp_op = VectorJacobianProduct(
            self.factor_out,
            fvjp,
            *self.args,
            out_shapes=fval.to_dict().shapes,
        )
        return fval, fvjp_op

    def _jvp_func_jacobian(
            self, values: VariableData, **kwargs
    ) ->  Tuple[FactorValue, "JacobianVectorProduct"]:
        args = (values[k] for k in self.args)
        raw_fval, raw_jac = self._factor_jacobian(*args, **kwargs)
        fval = self._factor_value(raw_fval)
        jvp = self._jac_out_to_jvp(raw_jac, values=fval.to_dict().merge(values))
        return fval, jvp

    func_jacobian = _jvp_func_jacobian

    def _factor_jacobian(self, *args) -> Tuple[Any, Any]:
        return self._factor_args(*args), self._jacobian(*args)

    def _factor_args(self, *args):
        return self._factor(*args)

    def _unpack_jacobian_out(self, raw_jac: Any) -> Dict[Variable, VariableData]:
        jac = {}
        for v0, vjac in nested_filter(is_variable, self.factor_out, raw_jac):
            jac[v0] = VariableData()
            for v1, j in zip(self.args, vjac):
                jac[v0][v1] = j

        return jac

    def _jac_out_to_jvp(
            self, raw_jac: Any, values: VariableData
    ) -> "JacobianVectorProduct":
        from autofit.graphical.factor_graphs.jacobians import (
            JacobianVectorProduct,
        )
        jac = self._unpack_jacobian_out(raw_jac)
        return JacobianVectorProduct.from_dense(jac, values=values)


class FactorKW(Factor):
    """Represents factors in Graphical models. The functions passed to this
    object will be called by keyword arguments

    Parameters
    ----------
    factor
        the function being wrapped, must accept calls
        through positional arguments
    **kwargs: Variables
        Variables for each keyword argument for the function
    factor_out:
        The output of the factor. This can just be `FactorValue`
        or can be a arbitrarily nested structure of lists, tuples and dicts
        e.g.
        >>> foo = lambda x, y: (z, {'a': [a]})
        >>> factor = Factor(foo, x, y, factor_out=(z, {'a': [a]}))
    name: optional, str
        the name of the factor, if not passed then uses the name
        of the function passed

    Methods
    -------
    __call__({x: x0}) -> FactorValue
        calls the factor, the passed input must be a dictionary with
        where the keys are the Variable objects that the function takes
        as input.

        returns a FactorValue object which behaves like an np.ndarray
        deterministic values are stored in the deterministic_values
        attribute

    func_jacobian({x: x0}) -> Tuple[FactorValue, AbstractJacobianValue]
        calls the factor and returns it value and the Jacobian of its value
        with respect to the `variables` passed. The Jacobian is returned as
        a VariableLinearOperator with the appropriate methods for calculating
        the vector-Jacobian or Jacobian-vector products depending on how
        the Jacobian is calculated internally.
    """

    def __init__(
            self,
            factor,
            name="",
            arg_names=None,
            factor_out=FactorValue,
            plates: Tuple[Plate, ...] = (),
            vjp=False,
            factor_vjp=None,
            factor_jacobian=None,
            jacobian=None,
            numerical_jacobian=True,
            jacfwd=True,
            eps=1e-8,
            **kwargs: Variable,
    ):
        args = tuple(kwargs.values())
        arg_names = tuple(kwargs.keys())
        super().__init__(
            factor,
            *args,
            name=name,
            arg_names=arg_names,
            factor_out=factor_out,
            plates=plates,
            vjp=vjp,
            factor_vjp=factor_vjp,
            factor_jacobian=factor_jacobian,
            jacobian=jacobian,
            numerical_jacobian=numerical_jacobian,
            jacfwd=jacfwd,
            eps=eps,
        )

    _factor_args = AbstractFactor._factor_args



import logging

from .analysis import Analysis
from .combined import CombinedAnalysis
from ..paths.abstract import AbstractPaths

logger = logging.getLogger(
    __name__
)


class IndexedAnalysis:
    def __init__(self, analysis: Analysis, index: int):
        """
        One instance in a collection corresponds to this analysis. That
        instance is identified by its index in the collection.

        Parameters
        ----------
        analysis
            An analysis that can be applied to an instance in a collection
        index
            The index of the instance that should be passed to the analysis
        """
        if isinstance(analysis, IndexedAnalysis):
            analysis = analysis.analysis
        self.analysis = analysis
        self.index = index

    def log_likelihood_function(self, instance):
        """
        Compute the log likelihood by taking the instance at the index
        """
        return self.analysis.log_likelihood_function(
            instance[self.index]
        )

    def visualize(self, paths: AbstractPaths, instance, during_analysis):
        return self.analysis.visualize(
            paths, instance[self.index], during_analysis
        )

    def profile_log_likelihood_function(self, paths: AbstractPaths, instance):
        return self.profile_log_likelihood_function(
            paths, instance[self.index]
        )

    def __getattr__(self, item):
        return getattr(
            self.analysis,
            item
        )

    def make_result(
            self,
            samples,
            model,
            sigma=3.0,
            use_errors=True,
            use_widths=True
    ):
        return self.analysis.make_result(
            samples,
            model,
            sigma=sigma,
            use_errors=use_errors,
            use_widths=use_widths,
        )


class IndexCollectionAnalysis(CombinedAnalysis):
    def __init__(self, *analyses):
        """
        Collection of analyses where each analysis has a different
        corresponding model.

        Parameters
        ----------
        analyses
            A list of analyses each with a separate model
        """
        super().__init__(*[
            IndexedAnalysis(
                analysis,
                index,
            )
            for index, analysis
            in enumerate(analyses)
        ])

    def make_result(
            self,
            samples,
            model,
            sigma=1.0,
            use_errors=True,
            use_widths=False
    ):
        """
        Associate each model with an analysis when creating the result.
        """
        child_results = [
            analysis.make_result(
                samples.subsamples(model),
                model,
                sigma=sigma,
                use_errors=use_errors,
                use_widths=use_widths,
            )
            for model, analysis in zip(model, self.analyses)
        ]
        result = self.analyses[0].make_result(
            samples=samples,
            model=model,
            sigma=sigma,
            use_errors=use_errors,
            use_widths=use_widths,
        )
        result.child_results = child_results
        return result



from abc import ABC
from os import path
from typing import Optional

import numpy as np

from autoconf import conf
from autofit.database.sqlalchemy_ import sa
from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.non_linear.abstract_search import PriorPasser
from autofit.non_linear.nest.abstract_nest import AbstractNest
from autofit.non_linear.nest.dynesty.plotter import DynestyPlotter
from autofit.non_linear.nest.dynesty.samples import DynestySamples
from autofit.plot.output import Output


def prior_transform(cube, model):
    phys_cube = model.vector_from_unit_vector(
        unit_vector=cube,
        ignore_prior_limits=True
    )

    for i in range(len(phys_cube)):
        cube[i] = phys_cube[i]

    return cube


class AbstractDynesty(AbstractNest, ABC):

    def __init__(
            self,
            name: str = "",
            path_prefix: str = "",
            unique_tag: Optional[str] = None,
            prior_passer: PriorPasser = None,
            iterations_per_update: int = None,
            number_of_cores: int = None,
            session: Optional[sa.orm.Session] = None,
            **kwargs
    ):
        """
        A Dynesty non-linear search.

        For a full description of Dynesty, checkout its GitHub and readthedocs webpages:

        https://github.com/joshspeagle/dynesty
        https://dynesty.readthedocs.io/en/latest/index.html

        Parameters
        ----------
        name
            The name of the search, controlling the last folder results are output.
        path_prefix
            The path of folders prefixing the name folder where results are output.
        unique_tag
            The name of a unique tag for this model-fit, which will be given a unique entry in the sqlite database
            and also acts as the folder after the path prefix and before the search name.
        prior_passer
            Controls how priors are passed from the results of this `NonLinearSearch` to a subsequent non-linear search.
        iterations_per_update
            The number of iterations performed between every Dynesty back-up (via dumping the Dynesty instance as a
            pickle).
        number_of_cores
            The number of cores Emcee sampling is performed using a Python multiprocessing Pool instance. If 1, a
            pool instance is not created and the job runs in serial.
        session
            An SQLalchemy session instance so the results of the model-fit are written to an SQLite database.
        """

        super().__init__(
            name=name,
            path_prefix=path_prefix,
            unique_tag=unique_tag,
            prior_passer=prior_passer,
            iterations_per_update=iterations_per_update,
            session=session,
            **kwargs
        )

        self.number_of_cores = (
            self._config("parallel", "number_of_cores")
            if number_of_cores is None
            else number_of_cores
        )

        self.logger.debug("Creating DynestyStatic Search")

    class Fitness(AbstractNest.Fitness):
        @property
        def resample_figure_of_merit(self):
            """
            If a sample raises a FitException, this value is returned to signify that the point requires resampling or
             should be given a likelihood so low that it is discard.

             -np.inf is an invalid sample value for Dynesty, so we instead use a large negative number."""
            return -1.0e99

    def _fit(self, model: AbstractPriorModel, analysis, log_likelihood_cap=None):
        """
        Fit a model using Dynesty and the Analysis class which contains the data and returns the log likelihood from
        instances of the model, which the `NonLinearSearch` seeks to maximize.

        Parameters
        ----------
        model : ModelMapper
            The model which generates instances for different points in parameter space.
        analysis : Analysis
            Contains the data and the log likelihood function which fits an instance of the model to the data, returning
            the log likelihood the `NonLinearSearch` maximizes.

        Returns
        -------
        A result object comprising the Samples object that includes the maximum log likelihood instance and full
        set of accepted samples of the fit.
        """

        fitness_function = self.fitness_function_from_model_and_analysis(
            model=model, analysis=analysis, log_likelihood_cap=log_likelihood_cap,
        )

        if self.paths.is_object("dynesty"):

            sampler = self.paths.load_object(
                "dynesty"
            )
            sampler.loglikelihood = fitness_function

            sampler.rstate = np.random

            if self.number_of_cores == 1:
                sampler.M = map
            else:
                pool = self.make_sneaky_pool(
                    fitness_function
                )

                sampler.M = pool.map
                sampler.pool = pool

            self.logger.info("Existing Dynesty samples found, resuming non-linear search.")

        else:
            if self.number_of_cores > 1:
                pool = self.make_sneaky_pool(
                    fitness_function
                )
            else:
                pool = None

            sampler = self.sampler_from(
                model=model,
                fitness_function=fitness_function,
                pool=pool
            )

            self.logger.info("No Dynesty samples found, beginning new non-linear search. ")

        if not hasattr(sampler, "pool"):
            sampler.pool = None

        finished = False

        while not finished:

            try:
                total_iterations = np.sum(sampler.results.ncall)
            except AttributeError:
                total_iterations = 0

            if self.config_dict_run["maxcall"] is not None:
                iterations = self.config_dict_run["maxcall"] - total_iterations
            else:
                iterations = self.iterations_per_update

            if iterations > 0:

                for i in range(10):

                    try:

                        config_dict_run = self.config_dict_run
                        config_dict_run.pop("maxcall")

                        sampler.run_nested(
                            maxcall=iterations,
                            print_progress=not self.silence,
                            **config_dict_run
                        )

                        if i == 9:
                            raise ValueError("Dynesty crashed due to repeated bounding errors")

                        break

                    except (ValueError, np.linalg.LinAlgError):

                        continue

            sampler.loglikelihood = None

            self.paths.save_object(
                "dynesty",
                sampler
            )

            sampler.loglikelihood = fitness_function

            self.perform_update(model=model, analysis=analysis, during_analysis=True)

            iterations_after_run = np.sum(sampler.results.ncall)

            if (
                    total_iterations == iterations_after_run
                    or total_iterations == self.config_dict_run["maxcall"]
            ):
                finished = True

    def sampler_from(
            self,
            model,
            fitness_function,
            pool=None
    ):
        return NotImplementedError()

    def samples_from(self, model):
        """
        Create a `Samples` object from this non-linear search's output files on the hard-disk and model.

        For Dynesty, all information that we need is available from the instance of the dynesty sampler.

        Parameters
        ----------
        model
            The model which generates instances for different points in parameter space. This maps the points from unit
            cube values to physical values via the priors.
        """
        sampler = self.paths.load_object(
            "dynesty"
        )

        return DynestySamples.from_results_internal(
            model=model,
            results_internal=sampler.results,
            number_live_points=self.total_live_points,
            unconverged_sample_size=1,
            time=self.timer.time,
        )

    def samples_via_results_from(self, model):
        """
        Create a `Samples` object from this non-linear search's output files on the hard-disk and model.

        For Dynesty, all information that we need is available from the instance of the dynesty sampler.

        Parameters
        ----------
        model
            The model which generates instances for different points in parameter space. This maps the points from unit
            cube values to physical values via the priors.
        """
        results_internal = self.paths.load_object(
            "results"
        )

        return DynestySamples.from_results_internal(
            model=model,
            results_internal=results_internal,
            number_live_points=self.total_live_points,
            unconverged_sample_size=1,
            time=self.timer.time,
        )

    def live_points_from_model_and_fitness_function(
            self, model, fitness_function
    ):

        unit_parameters, parameters, log_likelihood_list = self.initializer.samples_from_model(
            total_points=self.total_live_points,
            model=model,
            fitness_function=fitness_function,
        )

        init_unit_parameters = np.zeros(shape=(self.total_live_points, model.prior_count))
        init_parameters = np.zeros(shape=(self.total_live_points, model.prior_count))
        init_log_likelihood_list = np.zeros(shape=(self.total_live_points))

        for index in range(len(parameters)):
            init_unit_parameters[index, :] = np.asarray(unit_parameters[index])
            init_parameters[index, :] = np.asarray(parameters[index])
            init_log_likelihood_list[index] = np.asarray(log_likelihood_list[index])

        return [init_unit_parameters, init_parameters, init_log_likelihood_list]

    def remove_state_files(self):
        self.paths.remove_object("dynesty")

    @property
    def total_live_points(self):
        raise NotImplementedError()

    def plot_results(self, samples):

        if not samples.pdf_converged:
            return

        def should_plot(name):
            return conf.instance["visualize"]["plots_search"]["dynesty"][name]

        plotter = DynestyPlotter(
            samples=samples,
            output=Output(path=path.join(self.paths.image_path, "search"), format="png")
        )

        if should_plot("cornerplot"):
            plotter.cornerplot()

        if should_plot("runplot"):
            plotter.runplot()

        if should_plot("traceplot"):
            plotter.traceplot()

        if should_plot("cornerpoints"):
            plotter.cornerpoints()
